# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

from pkg_resources import parse_version
import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO
import collections
from enum import Enum


if parse_version(kaitaistruct.__version__) < parse_version('0.9'):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

class S3m(KaitaiStruct):
    """Scream Tracker 3 module is a tracker music file format that, as all
    tracker music, bundles both sound samples and instructions on which
    notes to play. It originates from a Scream Tracker 3 music editor
    (1994) by Future Crew, derived from original Scream Tracker 2 (.stm)
    module format.
    
    Instrument descriptions in S3M format allow to use either digital
    samples or setup and control AdLib (OPL2) synth.
    
    Music is organized in so called `patterns`. "Pattern" is a generally
    a 64-row long table, which instructs which notes to play on which
    time measure. "Patterns" are played one-by-one in a sequence
    determined by `orders`, which is essentially an array of pattern IDs
    - this way it's possible to reuse certain patterns more than once
    for repetitive musical phrases.
    
    .. seealso::
       Source - http://hackipedia.org/browse.cgi/File%20formats/Music%20tracker/S3M%2c%20ScreamTracker%203/Scream%20Tracker%203.20%20by%20Future%20Crew.txt
    """
    SEQ_FIELDS = ["song_name", "magic1", "file_type", "reserved1", "num_orders", "num_instruments", "num_patterns", "flags", "version", "samples_format", "magic2", "global_volume", "initial_speed", "initial_tempo", "is_stereo", "master_volume", "ultra_click_removal", "has_custom_pan", "reserved2", "ofs_special", "channels", "orders", "instruments", "patterns", "channel_pans"]
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._debug = collections.defaultdict(dict)

    def _read(self):
        self._debug['song_name']['start'] = self._io.pos()
        self.song_name = KaitaiStream.bytes_terminate(self._io.read_bytes(28), 0, False)
        self._debug['song_name']['end'] = self._io.pos()
        self._debug['magic1']['start'] = self._io.pos()
        self.magic1 = self._io.read_bytes(1)
        self._debug['magic1']['end'] = self._io.pos()
        if not self.magic1 == b"\x1A":
            raise kaitaistruct.ValidationNotEqualError(b"\x1A", self.magic1, self._io, u"/seq/1")
        self._debug['file_type']['start'] = self._io.pos()
        self.file_type = self._io.read_u1()
        self._debug['file_type']['end'] = self._io.pos()
        self._debug['reserved1']['start'] = self._io.pos()
        self.reserved1 = self._io.read_bytes(2)
        self._debug['reserved1']['end'] = self._io.pos()
        self._debug['num_orders']['start'] = self._io.pos()
        self.num_orders = self._io.read_u2le()
        self._debug['num_orders']['end'] = self._io.pos()
        self._debug['num_instruments']['start'] = self._io.pos()
        self.num_instruments = self._io.read_u2le()
        self._debug['num_instruments']['end'] = self._io.pos()
        self._debug['num_patterns']['start'] = self._io.pos()
        self.num_patterns = self._io.read_u2le()
        self._debug['num_patterns']['end'] = self._io.pos()
        self._debug['flags']['start'] = self._io.pos()
        self.flags = self._io.read_u2le()
        self._debug['flags']['end'] = self._io.pos()
        self._debug['version']['start'] = self._io.pos()
        self.version = self._io.read_u2le()
        self._debug['version']['end'] = self._io.pos()
        self._debug['samples_format']['start'] = self._io.pos()
        self.samples_format = self._io.read_u2le()
        self._debug['samples_format']['end'] = self._io.pos()
        self._debug['magic2']['start'] = self._io.pos()
        self.magic2 = self._io.read_bytes(4)
        self._debug['magic2']['end'] = self._io.pos()
        if not self.magic2 == b"\x53\x43\x52\x4D":
            raise kaitaistruct.ValidationNotEqualError(b"\x53\x43\x52\x4D", self.magic2, self._io, u"/seq/10")
        self._debug['global_volume']['start'] = self._io.pos()
        self.global_volume = self._io.read_u1()
        self._debug['global_volume']['end'] = self._io.pos()
        self._debug['initial_speed']['start'] = self._io.pos()
        self.initial_speed = self._io.read_u1()
        self._debug['initial_speed']['end'] = self._io.pos()
        self._debug['initial_tempo']['start'] = self._io.pos()
        self.initial_tempo = self._io.read_u1()
        self._debug['initial_tempo']['end'] = self._io.pos()
        self._debug['is_stereo']['start'] = self._io.pos()
        self.is_stereo = self._io.read_bits_int_be(1) != 0
        self._debug['is_stereo']['end'] = self._io.pos()
        self._debug['master_volume']['start'] = self._io.pos()
        self.master_volume = self._io.read_bits_int_be(7)
        self._debug['master_volume']['end'] = self._io.pos()
        self._io.align_to_byte()
        self._debug['ultra_click_removal']['start'] = self._io.pos()
        self.ultra_click_removal = self._io.read_u1()
        self._debug['ultra_click_removal']['end'] = self._io.pos()
        self._debug['has_custom_pan']['start'] = self._io.pos()
        self.has_custom_pan = self._io.read_u1()
        self._debug['has_custom_pan']['end'] = self._io.pos()
        self._debug['reserved2']['start'] = self._io.pos()
        self.reserved2 = self._io.read_bytes(8)
        self._debug['reserved2']['end'] = self._io.pos()
        self._debug['ofs_special']['start'] = self._io.pos()
        self.ofs_special = self._io.read_u2le()
        self._debug['ofs_special']['end'] = self._io.pos()
        self._debug['channels']['start'] = self._io.pos()
        self.channels = [None] * (32)
        for i in range(32):
            if not 'arr' in self._debug['channels']:
                self._debug['channels']['arr'] = []
            self._debug['channels']['arr'].append({'start': self._io.pos()})
            _t_channels = S3m.Channel(self._io, self, self._root)
            _t_channels._read()
            self.channels[i] = _t_channels
            self._debug['channels']['arr'][i]['end'] = self._io.pos()

        self._debug['channels']['end'] = self._io.pos()
        self._debug['orders']['start'] = self._io.pos()
        self.orders = self._io.read_bytes(self.num_orders)
        self._debug['orders']['end'] = self._io.pos()
        self._debug['instruments']['start'] = self._io.pos()
        self.instruments = [None] * (self.num_instruments)
        for i in range(self.num_instruments):
            if not 'arr' in self._debug['instruments']:
                self._debug['instruments']['arr'] = []
            self._debug['instruments']['arr'].append({'start': self._io.pos()})
            _t_instruments = S3m.InstrumentPtr(self._io, self, self._root)
            _t_instruments._read()
            self.instruments[i] = _t_instruments
            self._debug['instruments']['arr'][i]['end'] = self._io.pos()

        self._debug['instruments']['end'] = self._io.pos()
        self._debug['patterns']['start'] = self._io.pos()
        self.patterns = [None] * (self.num_patterns)
        for i in range(self.num_patterns):
            if not 'arr' in self._debug['patterns']:
                self._debug['patterns']['arr'] = []
            self._debug['patterns']['arr'].append({'start': self._io.pos()})
            _t_patterns = S3m.PatternPtr(self._io, self, self._root)
            _t_patterns._read()
            self.patterns[i] = _t_patterns
            self._debug['patterns']['arr'][i]['end'] = self._io.pos()

        self._debug['patterns']['end'] = self._io.pos()
        if self.has_custom_pan == 252:
            self._debug['channel_pans']['start'] = self._io.pos()
            self.channel_pans = [None] * (32)
            for i in range(32):
                if not 'arr' in self._debug['channel_pans']:
                    self._debug['channel_pans']['arr'] = []
                self._debug['channel_pans']['arr'].append({'start': self._io.pos()})
                _t_channel_pans = S3m.ChannelPan(self._io, self, self._root)
                _t_channel_pans._read()
                self.channel_pans[i] = _t_channel_pans
                self._debug['channel_pans']['arr'][i]['end'] = self._io.pos()

            self._debug['channel_pans']['end'] = self._io.pos()


    class ChannelPan(KaitaiStruct):
        SEQ_FIELDS = ["reserved1", "has_custom_pan", "reserved2", "pan"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['reserved1']['start'] = self._io.pos()
            self.reserved1 = self._io.read_bits_int_be(2)
            self._debug['reserved1']['end'] = self._io.pos()
            self._debug['has_custom_pan']['start'] = self._io.pos()
            self.has_custom_pan = self._io.read_bits_int_be(1) != 0
            self._debug['has_custom_pan']['end'] = self._io.pos()
            self._debug['reserved2']['start'] = self._io.pos()
            self.reserved2 = self._io.read_bits_int_be(1) != 0
            self._debug['reserved2']['end'] = self._io.pos()
            self._debug['pan']['start'] = self._io.pos()
            self.pan = self._io.read_bits_int_be(4)
            self._debug['pan']['end'] = self._io.pos()


    class PatternCell(KaitaiStruct):
        SEQ_FIELDS = ["has_fx", "has_volume", "has_note_and_instrument", "channel_num", "note", "instrument", "volume", "fx_type", "fx_value"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['has_fx']['start'] = self._io.pos()
            self.has_fx = self._io.read_bits_int_be(1) != 0
            self._debug['has_fx']['end'] = self._io.pos()
            self._debug['has_volume']['start'] = self._io.pos()
            self.has_volume = self._io.read_bits_int_be(1) != 0
            self._debug['has_volume']['end'] = self._io.pos()
            self._debug['has_note_and_instrument']['start'] = self._io.pos()
            self.has_note_and_instrument = self._io.read_bits_int_be(1) != 0
            self._debug['has_note_and_instrument']['end'] = self._io.pos()
            self._debug['channel_num']['start'] = self._io.pos()
            self.channel_num = self._io.read_bits_int_be(5)
            self._debug['channel_num']['end'] = self._io.pos()
            self._io.align_to_byte()
            if self.has_note_and_instrument:
                self._debug['note']['start'] = self._io.pos()
                self.note = self._io.read_u1()
                self._debug['note']['end'] = self._io.pos()

            if self.has_note_and_instrument:
                self._debug['instrument']['start'] = self._io.pos()
                self.instrument = self._io.read_u1()
                self._debug['instrument']['end'] = self._io.pos()

            if self.has_volume:
                self._debug['volume']['start'] = self._io.pos()
                self.volume = self._io.read_u1()
                self._debug['volume']['end'] = self._io.pos()

            if self.has_fx:
                self._debug['fx_type']['start'] = self._io.pos()
                self.fx_type = self._io.read_u1()
                self._debug['fx_type']['end'] = self._io.pos()

            if self.has_fx:
                self._debug['fx_value']['start'] = self._io.pos()
                self.fx_value = self._io.read_u1()
                self._debug['fx_value']['end'] = self._io.pos()



    class PatternCells(KaitaiStruct):
        SEQ_FIELDS = ["cells"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['cells']['start'] = self._io.pos()
            self.cells = []
            i = 0
            while not self._io.is_eof():
                if not 'arr' in self._debug['cells']:
                    self._debug['cells']['arr'] = []
                self._debug['cells']['arr'].append({'start': self._io.pos()})
                _t_cells = S3m.PatternCell(self._io, self, self._root)
                _t_cells._read()
                self.cells.append(_t_cells)
                self._debug['cells']['arr'][len(self.cells) - 1]['end'] = self._io.pos()
                i += 1

            self._debug['cells']['end'] = self._io.pos()


    class Channel(KaitaiStruct):
        SEQ_FIELDS = ["is_disabled", "ch_type"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['is_disabled']['start'] = self._io.pos()
            self.is_disabled = self._io.read_bits_int_be(1) != 0
            self._debug['is_disabled']['end'] = self._io.pos()
            self._debug['ch_type']['start'] = self._io.pos()
            self.ch_type = self._io.read_bits_int_be(7)
            self._debug['ch_type']['end'] = self._io.pos()


    class SwappedU3(KaitaiStruct):
        """Custom 3-byte integer, stored in mixed endian manner."""
        SEQ_FIELDS = ["hi", "lo"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['hi']['start'] = self._io.pos()
            self.hi = self._io.read_u1()
            self._debug['hi']['end'] = self._io.pos()
            self._debug['lo']['start'] = self._io.pos()
            self.lo = self._io.read_u2le()
            self._debug['lo']['end'] = self._io.pos()

        @property
        def value(self):
            if hasattr(self, '_m_value'):
                return self._m_value if hasattr(self, '_m_value') else None

            self._m_value = (self.lo | (self.hi << 16))
            return self._m_value if hasattr(self, '_m_value') else None


    class Pattern(KaitaiStruct):
        SEQ_FIELDS = ["size", "body"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['size']['start'] = self._io.pos()
            self.size = self._io.read_u2le()
            self._debug['size']['end'] = self._io.pos()
            self._debug['body']['start'] = self._io.pos()
            self._raw_body = self._io.read_bytes((self.size - 2))
            _io__raw_body = KaitaiStream(BytesIO(self._raw_body))
            self.body = S3m.PatternCells(_io__raw_body, self, self._root)
            self.body._read()
            self._debug['body']['end'] = self._io.pos()


    class PatternPtr(KaitaiStruct):
        SEQ_FIELDS = ["paraptr"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['paraptr']['start'] = self._io.pos()
            self.paraptr = self._io.read_u2le()
            self._debug['paraptr']['end'] = self._io.pos()

        @property
        def body(self):
            if hasattr(self, '_m_body'):
                return self._m_body if hasattr(self, '_m_body') else None

            _pos = self._io.pos()
            self._io.seek((self.paraptr * 16))
            self._debug['_m_body']['start'] = self._io.pos()
            self._m_body = S3m.Pattern(self._io, self, self._root)
            self._m_body._read()
            self._debug['_m_body']['end'] = self._io.pos()
            self._io.seek(_pos)
            return self._m_body if hasattr(self, '_m_body') else None


    class InstrumentPtr(KaitaiStruct):
        SEQ_FIELDS = ["paraptr"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['paraptr']['start'] = self._io.pos()
            self.paraptr = self._io.read_u2le()
            self._debug['paraptr']['end'] = self._io.pos()

        @property
        def body(self):
            if hasattr(self, '_m_body'):
                return self._m_body if hasattr(self, '_m_body') else None

            _pos = self._io.pos()
            self._io.seek((self.paraptr * 16))
            self._debug['_m_body']['start'] = self._io.pos()
            self._m_body = S3m.Instrument(self._io, self, self._root)
            self._m_body._read()
            self._debug['_m_body']['end'] = self._io.pos()
            self._io.seek(_pos)
            return self._m_body if hasattr(self, '_m_body') else None


    class Instrument(KaitaiStruct):

        class InstTypes(Enum):
            sample = 1
            melodic = 2
            bass_drum = 3
            snare_drum = 4
            tom = 5
            cymbal = 6
            hihat = 7
        SEQ_FIELDS = ["type", "filename", "body", "tuning_hz", "reserved2", "sample_name", "magic"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['type']['start'] = self._io.pos()
            self.type = KaitaiStream.resolve_enum(S3m.Instrument.InstTypes, self._io.read_u1())
            self._debug['type']['end'] = self._io.pos()
            self._debug['filename']['start'] = self._io.pos()
            self.filename = KaitaiStream.bytes_terminate(self._io.read_bytes(12), 0, False)
            self._debug['filename']['end'] = self._io.pos()
            self._debug['body']['start'] = self._io.pos()
            _on = self.type
            if _on == S3m.Instrument.InstTypes.sample:
                self.body = S3m.Instrument.Sampled(self._io, self, self._root)
                self.body._read()
            else:
                self.body = S3m.Instrument.Adlib(self._io, self, self._root)
                self.body._read()
            self._debug['body']['end'] = self._io.pos()
            self._debug['tuning_hz']['start'] = self._io.pos()
            self.tuning_hz = self._io.read_u4le()
            self._debug['tuning_hz']['end'] = self._io.pos()
            self._debug['reserved2']['start'] = self._io.pos()
            self.reserved2 = self._io.read_bytes(12)
            self._debug['reserved2']['end'] = self._io.pos()
            self._debug['sample_name']['start'] = self._io.pos()
            self.sample_name = KaitaiStream.bytes_terminate(self._io.read_bytes(28), 0, False)
            self._debug['sample_name']['end'] = self._io.pos()
            self._debug['magic']['start'] = self._io.pos()
            self.magic = self._io.read_bytes(4)
            self._debug['magic']['end'] = self._io.pos()
            if not self.magic == b"\x53\x43\x52\x53":
                raise kaitaistruct.ValidationNotEqualError(b"\x53\x43\x52\x53", self.magic, self._io, u"/types/instrument/seq/6")

        class Sampled(KaitaiStruct):
            SEQ_FIELDS = ["paraptr_sample", "len_sample", "loop_begin", "loop_end", "default_volume", "reserved1", "is_packed", "flags"]
            def __init__(self, _io, _parent=None, _root=None):
                self._io = _io
                self._parent = _parent
                self._root = _root if _root else self
                self._debug = collections.defaultdict(dict)

            def _read(self):
                self._debug['paraptr_sample']['start'] = self._io.pos()
                self.paraptr_sample = S3m.SwappedU3(self._io, self, self._root)
                self.paraptr_sample._read()
                self._debug['paraptr_sample']['end'] = self._io.pos()
                self._debug['len_sample']['start'] = self._io.pos()
                self.len_sample = self._io.read_u4le()
                self._debug['len_sample']['end'] = self._io.pos()
                self._debug['loop_begin']['start'] = self._io.pos()
                self.loop_begin = self._io.read_u4le()
                self._debug['loop_begin']['end'] = self._io.pos()
                self._debug['loop_end']['start'] = self._io.pos()
                self.loop_end = self._io.read_u4le()
                self._debug['loop_end']['end'] = self._io.pos()
                self._debug['default_volume']['start'] = self._io.pos()
                self.default_volume = self._io.read_u1()
                self._debug['default_volume']['end'] = self._io.pos()
                self._debug['reserved1']['start'] = self._io.pos()
                self.reserved1 = self._io.read_u1()
                self._debug['reserved1']['end'] = self._io.pos()
                self._debug['is_packed']['start'] = self._io.pos()
                self.is_packed = self._io.read_u1()
                self._debug['is_packed']['end'] = self._io.pos()
                self._debug['flags']['start'] = self._io.pos()
                self.flags = self._io.read_u1()
                self._debug['flags']['end'] = self._io.pos()

            @property
            def sample(self):
                if hasattr(self, '_m_sample'):
                    return self._m_sample if hasattr(self, '_m_sample') else None

                _pos = self._io.pos()
                self._io.seek((self.paraptr_sample.value * 16))
                self._debug['_m_sample']['start'] = self._io.pos()
                self._m_sample = self._io.read_bytes(self.len_sample)
                self._debug['_m_sample']['end'] = self._io.pos()
                self._io.seek(_pos)
                return self._m_sample if hasattr(self, '_m_sample') else None


        class Adlib(KaitaiStruct):
            SEQ_FIELDS = ["reserved1", "_unnamed1"]
            def __init__(self, _io, _parent=None, _root=None):
                self._io = _io
                self._parent = _parent
                self._root = _root if _root else self
                self._debug = collections.defaultdict(dict)

            def _read(self):
                self._debug['reserved1']['start'] = self._io.pos()
                self.reserved1 = self._io.read_bytes(3)
                self._debug['reserved1']['end'] = self._io.pos()
                if not self.reserved1 == b"\x00\x00\x00":
                    raise kaitaistruct.ValidationNotEqualError(b"\x00\x00\x00", self.reserved1, self._io, u"/types/instrument/types/adlib/seq/0")
                self._debug['_unnamed1']['start'] = self._io.pos()
                self._unnamed1 = self._io.read_bytes(16)
                self._debug['_unnamed1']['end'] = self._io.pos()







# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

from pkg_resources import parse_version
import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO
from enum import Enum
import collections


if parse_version(kaitaistruct.__version__) < parse_version('0.9'):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

class Riff(KaitaiStruct):
    """The Resource Interchange File Format (RIFF) is a generic file container format
    for storing data in tagged chunks. It is primarily used to store multimedia
    such as sound and video, though it may also be used to store any arbitrary data.
    
    The Microsoft implementation is mostly known through container formats
    like AVI, ANI and WAV, which use RIFF as their basis.
    
    .. seealso::
       Source - https://www.johnloomis.org/cpe102/asgn/asgn1/riff.html
    """

    class Fourcc(Enum):
        riff = 1179011410
        info = 1330007625
        list = 1414744396
    SEQ_FIELDS = ["chunk"]
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._debug = collections.defaultdict(dict)

    def _read(self):
        self._debug['chunk']['start'] = self._io.pos()
        self.chunk = Riff.Chunk(self._io, self, self._root)
        self.chunk._read()
        self._debug['chunk']['end'] = self._io.pos()

    class ListChunkData(KaitaiStruct):
        SEQ_FIELDS = ["save_parent_chunk_data_ofs", "parent_chunk_data"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            if self.parent_chunk_data_ofs < 0:
                self._debug['save_parent_chunk_data_ofs']['start'] = self._io.pos()
                self.save_parent_chunk_data_ofs = self._io.read_bytes(0)
                self._debug['save_parent_chunk_data_ofs']['end'] = self._io.pos()

            self._debug['parent_chunk_data']['start'] = self._io.pos()
            self.parent_chunk_data = Riff.ParentChunkData(self._io, self, self._root)
            self.parent_chunk_data._read()
            self._debug['parent_chunk_data']['end'] = self._io.pos()

        @property
        def parent_chunk_data_ofs(self):
            if hasattr(self, '_m_parent_chunk_data_ofs'):
                return self._m_parent_chunk_data_ofs if hasattr(self, '_m_parent_chunk_data_ofs') else None

            self._m_parent_chunk_data_ofs = self._io.pos()
            return self._m_parent_chunk_data_ofs if hasattr(self, '_m_parent_chunk_data_ofs') else None

        @property
        def form_type(self):
            if hasattr(self, '_m_form_type'):
                return self._m_form_type if hasattr(self, '_m_form_type') else None

            self._m_form_type = KaitaiStream.resolve_enum(Riff.Fourcc, self.parent_chunk_data.form_type)
            return self._m_form_type if hasattr(self, '_m_form_type') else None

        @property
        def form_type_readable(self):
            if hasattr(self, '_m_form_type_readable'):
                return self._m_form_type_readable if hasattr(self, '_m_form_type_readable') else None

            _pos = self._io.pos()
            self._io.seek(self.parent_chunk_data_ofs)
            self._debug['_m_form_type_readable']['start'] = self._io.pos()
            self._m_form_type_readable = (self._io.read_bytes(4)).decode(u"ASCII")
            self._debug['_m_form_type_readable']['end'] = self._io.pos()
            self._io.seek(_pos)
            return self._m_form_type_readable if hasattr(self, '_m_form_type_readable') else None

        @property
        def subchunks(self):
            if hasattr(self, '_m_subchunks'):
                return self._m_subchunks if hasattr(self, '_m_subchunks') else None

            io = self.parent_chunk_data.subchunks_slot._io
            _pos = io.pos()
            io.seek(0)
            self._debug['_m_subchunks']['start'] = io.pos()
            self._m_subchunks = []
            i = 0
            while not io.is_eof():
                if not 'arr' in self._debug['_m_subchunks']:
                    self._debug['_m_subchunks']['arr'] = []
                self._debug['_m_subchunks']['arr'].append({'start': io.pos()})
                _on = self.form_type
                if _on == Riff.Fourcc.info:
                    if not 'arr' in self._debug['_m_subchunks']:
                        self._debug['_m_subchunks']['arr'] = []
                    self._debug['_m_subchunks']['arr'].append({'start': io.pos()})
                    _t__m_subchunks = Riff.InfoSubchunk(io, self, self._root)
                    _t__m_subchunks._read()
                    self._m_subchunks.append(_t__m_subchunks)
                    self._debug['_m_subchunks']['arr'][len(self._m_subchunks) - 1]['end'] = io.pos()
                else:
                    if not 'arr' in self._debug['_m_subchunks']:
                        self._debug['_m_subchunks']['arr'] = []
                    self._debug['_m_subchunks']['arr'].append({'start': io.pos()})
                    _t__m_subchunks = Riff.ChunkType(io, self, self._root)
                    _t__m_subchunks._read()
                    self._m_subchunks.append(_t__m_subchunks)
                    self._debug['_m_subchunks']['arr'][len(self._m_subchunks) - 1]['end'] = io.pos()
                self._debug['_m_subchunks']['arr'][len(self._m_subchunks) - 1]['end'] = io.pos()
                i += 1

            self._debug['_m_subchunks']['end'] = io.pos()
            io.seek(_pos)
            return self._m_subchunks if hasattr(self, '_m_subchunks') else None


    class Chunk(KaitaiStruct):
        SEQ_FIELDS = ["id", "len", "data_slot", "pad_byte"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['id']['start'] = self._io.pos()
            self.id = self._io.read_u4le()
            self._debug['id']['end'] = self._io.pos()
            self._debug['len']['start'] = self._io.pos()
            self.len = self._io.read_u4le()
            self._debug['len']['end'] = self._io.pos()
            self._debug['data_slot']['start'] = self._io.pos()
            self._raw_data_slot = self._io.read_bytes(self.len)
            _io__raw_data_slot = KaitaiStream(BytesIO(self._raw_data_slot))
            self.data_slot = Riff.Chunk.Slot(_io__raw_data_slot, self, self._root)
            self.data_slot._read()
            self._debug['data_slot']['end'] = self._io.pos()
            self._debug['pad_byte']['start'] = self._io.pos()
            self.pad_byte = self._io.read_bytes((self.len % 2))
            self._debug['pad_byte']['end'] = self._io.pos()

        class Slot(KaitaiStruct):
            SEQ_FIELDS = []
            def __init__(self, _io, _parent=None, _root=None):
                self._io = _io
                self._parent = _parent
                self._root = _root if _root else self
                self._debug = collections.defaultdict(dict)

            def _read(self):
                pass



    class ParentChunkData(KaitaiStruct):
        SEQ_FIELDS = ["form_type", "subchunks_slot"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['form_type']['start'] = self._io.pos()
            self.form_type = self._io.read_u4le()
            self._debug['form_type']['end'] = self._io.pos()
            self._debug['subchunks_slot']['start'] = self._io.pos()
            self._raw_subchunks_slot = self._io.read_bytes_full()
            _io__raw_subchunks_slot = KaitaiStream(BytesIO(self._raw_subchunks_slot))
            self.subchunks_slot = Riff.ParentChunkData.Slot(_io__raw_subchunks_slot, self, self._root)
            self.subchunks_slot._read()
            self._debug['subchunks_slot']['end'] = self._io.pos()

        class Slot(KaitaiStruct):
            SEQ_FIELDS = []
            def __init__(self, _io, _parent=None, _root=None):
                self._io = _io
                self._parent = _parent
                self._root = _root if _root else self
                self._debug = collections.defaultdict(dict)

            def _read(self):
                pass



    class InfoSubchunk(KaitaiStruct):
        """All registered subchunks in the INFO chunk are NULL-terminated strings,
        but the unregistered might not be. By convention, the registered
        chunk IDs are in uppercase and the unregistered IDs are in lowercase.
        
        If the chunk ID of an INFO subchunk contains a lowercase
        letter, this chunk is considered as unregistered and thus we can make
        no assumptions about the type of data.
        """
        SEQ_FIELDS = ["save_chunk_ofs", "chunk"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            if self.chunk_ofs < 0:
                self._debug['save_chunk_ofs']['start'] = self._io.pos()
                self.save_chunk_ofs = self._io.read_bytes(0)
                self._debug['save_chunk_ofs']['end'] = self._io.pos()

            self._debug['chunk']['start'] = self._io.pos()
            self.chunk = Riff.Chunk(self._io, self, self._root)
            self.chunk._read()
            self._debug['chunk']['end'] = self._io.pos()

        @property
        def chunk_data(self):
            if hasattr(self, '_m_chunk_data'):
                return self._m_chunk_data if hasattr(self, '_m_chunk_data') else None

            io = self.chunk.data_slot._io
            _pos = io.pos()
            io.seek(0)
            self._debug['_m_chunk_data']['start'] = io.pos()
            _on = self.is_unregistered_tag
            if _on == False:
                self._m_chunk_data = (io.read_bytes_term(0, False, True, True)).decode(u"UTF-8")
            self._debug['_m_chunk_data']['end'] = io.pos()
            io.seek(_pos)
            return self._m_chunk_data if hasattr(self, '_m_chunk_data') else None

        @property
        def is_unregistered_tag(self):
            """Check if chunk_id contains lowercase characters ([a-z], ASCII 97 = a, ASCII 122 = z).
            """
            if hasattr(self, '_m_is_unregistered_tag'):
                return self._m_is_unregistered_tag if hasattr(self, '_m_is_unregistered_tag') else None

            self._m_is_unregistered_tag =  (( ((KaitaiStream.byte_array_index(self.id_chars, 0) >= 97) and (KaitaiStream.byte_array_index(self.id_chars, 0) <= 122)) ) or ( ((KaitaiStream.byte_array_index(self.id_chars, 1) >= 97) and (KaitaiStream.byte_array_index(self.id_chars, 1) <= 122)) ) or ( ((KaitaiStream.byte_array_index(self.id_chars, 2) >= 97) and (KaitaiStream.byte_array_index(self.id_chars, 2) <= 122)) ) or ( ((KaitaiStream.byte_array_index(self.id_chars, 3) >= 97) and (KaitaiStream.byte_array_index(self.id_chars, 3) <= 122)) )) 
            return self._m_is_unregistered_tag if hasattr(self, '_m_is_unregistered_tag') else None

        @property
        def id_chars(self):
            if hasattr(self, '_m_id_chars'):
                return self._m_id_chars if hasattr(self, '_m_id_chars') else None

            _pos = self._io.pos()
            self._io.seek(self.chunk_ofs)
            self._debug['_m_id_chars']['start'] = self._io.pos()
            self._m_id_chars = self._io.read_bytes(4)
            self._debug['_m_id_chars']['end'] = self._io.pos()
            self._io.seek(_pos)
            return self._m_id_chars if hasattr(self, '_m_id_chars') else None

        @property
        def chunk_id_readable(self):
            if hasattr(self, '_m_chunk_id_readable'):
                return self._m_chunk_id_readable if hasattr(self, '_m_chunk_id_readable') else None

            self._m_chunk_id_readable = (self.id_chars).decode(u"ASCII")
            return self._m_chunk_id_readable if hasattr(self, '_m_chunk_id_readable') else None

        @property
        def chunk_ofs(self):
            if hasattr(self, '_m_chunk_ofs'):
                return self._m_chunk_ofs if hasattr(self, '_m_chunk_ofs') else None

            self._m_chunk_ofs = self._io.pos()
            return self._m_chunk_ofs if hasattr(self, '_m_chunk_ofs') else None


    class ChunkType(KaitaiStruct):
        SEQ_FIELDS = ["save_chunk_ofs", "chunk"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            if self.chunk_ofs < 0:
                self._debug['save_chunk_ofs']['start'] = self._io.pos()
                self.save_chunk_ofs = self._io.read_bytes(0)
                self._debug['save_chunk_ofs']['end'] = self._io.pos()

            self._debug['chunk']['start'] = self._io.pos()
            self.chunk = Riff.Chunk(self._io, self, self._root)
            self.chunk._read()
            self._debug['chunk']['end'] = self._io.pos()

        @property
        def chunk_ofs(self):
            if hasattr(self, '_m_chunk_ofs'):
                return self._m_chunk_ofs if hasattr(self, '_m_chunk_ofs') else None

            self._m_chunk_ofs = self._io.pos()
            return self._m_chunk_ofs if hasattr(self, '_m_chunk_ofs') else None

        @property
        def chunk_id(self):
            if hasattr(self, '_m_chunk_id'):
                return self._m_chunk_id if hasattr(self, '_m_chunk_id') else None

            self._m_chunk_id = KaitaiStream.resolve_enum(Riff.Fourcc, self.chunk.id)
            return self._m_chunk_id if hasattr(self, '_m_chunk_id') else None

        @property
        def chunk_id_readable(self):
            if hasattr(self, '_m_chunk_id_readable'):
                return self._m_chunk_id_readable if hasattr(self, '_m_chunk_id_readable') else None

            _pos = self._io.pos()
            self._io.seek(self.chunk_ofs)
            self._debug['_m_chunk_id_readable']['start'] = self._io.pos()
            self._m_chunk_id_readable = (self._io.read_bytes(4)).decode(u"ASCII")
            self._debug['_m_chunk_id_readable']['end'] = self._io.pos()
            self._io.seek(_pos)
            return self._m_chunk_id_readable if hasattr(self, '_m_chunk_id_readable') else None

        @property
        def chunk_data(self):
            if hasattr(self, '_m_chunk_data'):
                return self._m_chunk_data if hasattr(self, '_m_chunk_data') else None

            io = self.chunk.data_slot._io
            _pos = io.pos()
            io.seek(0)
            self._debug['_m_chunk_data']['start'] = io.pos()
            _on = self.chunk_id
            if _on == Riff.Fourcc.list:
                self._m_chunk_data = Riff.ListChunkData(io, self, self._root)
                self._m_chunk_data._read()
            self._debug['_m_chunk_data']['end'] = io.pos()
            io.seek(_pos)
            return self._m_chunk_data if hasattr(self, '_m_chunk_data') else None


    @property
    def chunk_id(self):
        if hasattr(self, '_m_chunk_id'):
            return self._m_chunk_id if hasattr(self, '_m_chunk_id') else None

        self._m_chunk_id = KaitaiStream.resolve_enum(Riff.Fourcc, self.chunk.id)
        return self._m_chunk_id if hasattr(self, '_m_chunk_id') else None

    @property
    def is_riff_chunk(self):
        if hasattr(self, '_m_is_riff_chunk'):
            return self._m_is_riff_chunk if hasattr(self, '_m_is_riff_chunk') else None

        self._m_is_riff_chunk = self.chunk_id == Riff.Fourcc.riff
        return self._m_is_riff_chunk if hasattr(self, '_m_is_riff_chunk') else None

    @property
    def parent_chunk_data(self):
        if hasattr(self, '_m_parent_chunk_data'):
            return self._m_parent_chunk_data if hasattr(self, '_m_parent_chunk_data') else None

        if self.is_riff_chunk:
            io = self.chunk.data_slot._io
            _pos = io.pos()
            io.seek(0)
            self._debug['_m_parent_chunk_data']['start'] = io.pos()
            self._m_parent_chunk_data = Riff.ParentChunkData(io, self, self._root)
            self._m_parent_chunk_data._read()
            self._debug['_m_parent_chunk_data']['end'] = io.pos()
            io.seek(_pos)

        return self._m_parent_chunk_data if hasattr(self, '_m_parent_chunk_data') else None

    @property
    def subchunks(self):
        if hasattr(self, '_m_subchunks'):
            return self._m_subchunks if hasattr(self, '_m_subchunks') else None

        if self.is_riff_chunk:
            io = self.parent_chunk_data.subchunks_slot._io
            _pos = io.pos()
            io.seek(0)
            self._debug['_m_subchunks']['start'] = io.pos()
            self._m_subchunks = []
            i = 0
            while not io.is_eof():
                if not 'arr' in self._debug['_m_subchunks']:
                    self._debug['_m_subchunks']['arr'] = []
                self._debug['_m_subchunks']['arr'].append({'start': io.pos()})
                _t__m_subchunks = Riff.ChunkType(io, self, self._root)
                _t__m_subchunks._read()
                self._m_subchunks.append(_t__m_subchunks)
                self._debug['_m_subchunks']['arr'][len(self._m_subchunks) - 1]['end'] = io.pos()
                i += 1

            self._debug['_m_subchunks']['end'] = io.pos()
            io.seek(_pos)

        return self._m_subchunks if hasattr(self, '_m_subchunks') else None





# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

from pkg_resources import parse_version
import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO
import collections
from enum import Enum


if parse_version(kaitaistruct.__version__) < parse_version('0.9'):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

class SomeIpSdOptions(KaitaiStruct):
    """FormatOptions are used to transport additional information to the entries.
    This includes forinstance the information how a service instance is
    reachable (IP-Address, TransportProtocol, Port Number).
    
    .. seealso::
       section 4.1.2.4 Options Format - https://www.autosar.org/fileadmin/user_upload/standards/foundation/19-11/AUTOSAR_PRS_SOMEIPServiceDiscoveryProtocol.pdf
       -
    """
    SEQ_FIELDS = ["entries"]
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._debug = collections.defaultdict(dict)

    def _read(self):
        self._debug['entries']['start'] = self._io.pos()
        self.entries = []
        i = 0
        while not self._io.is_eof():
            if not 'arr' in self._debug['entries']:
                self._debug['entries']['arr'] = []
            self._debug['entries']['arr'].append({'start': self._io.pos()})
            _t_entries = SomeIpSdOptions.SdOption(self._io, self, self._root)
            _t_entries._read()
            self.entries.append(_t_entries)
            self._debug['entries']['arr'][len(self.entries) - 1]['end'] = self._io.pos()
            i += 1

        self._debug['entries']['end'] = self._io.pos()

    class SdOption(KaitaiStruct):

        class OptionTypes(Enum):
            configuration_option = 1
            load_balancing_option = 2
            ipv4_endpoint_option = 4
            ipv6_endpoint_option = 6
            ipv4_multicast_option = 20
            ipv6_multicast_option = 22
            ipv4_sd_endpoint_option = 36
            ipv6_sd_endpoint_option = 38
        SEQ_FIELDS = ["header", "content"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['header']['start'] = self._io.pos()
            self.header = SomeIpSdOptions.SdOption.SdOptionHeader(self._io, self, self._root)
            self.header._read()
            self._debug['header']['end'] = self._io.pos()
            self._debug['content']['start'] = self._io.pos()
            _on = self.header.type
            if _on == SomeIpSdOptions.SdOption.OptionTypes.load_balancing_option:
                self.content = SomeIpSdOptions.SdOption.SdLoadBalancingOption(self._io, self, self._root)
                self.content._read()
            elif _on == SomeIpSdOptions.SdOption.OptionTypes.configuration_option:
                self.content = SomeIpSdOptions.SdOption.SdConfigurationOption(self._io, self, self._root)
                self.content._read()
            elif _on == SomeIpSdOptions.SdOption.OptionTypes.ipv4_sd_endpoint_option:
                self.content = SomeIpSdOptions.SdOption.SdIpv4SdEndpointOption(self._io, self, self._root)
                self.content._read()
            elif _on == SomeIpSdOptions.SdOption.OptionTypes.ipv4_endpoint_option:
                self.content = SomeIpSdOptions.SdOption.SdIpv4EndpointOption(self._io, self, self._root)
                self.content._read()
            elif _on == SomeIpSdOptions.SdOption.OptionTypes.ipv6_sd_endpoint_option:
                self.content = SomeIpSdOptions.SdOption.SdIpv6SdEndpointOption(self._io, self, self._root)
                self.content._read()
            elif _on == SomeIpSdOptions.SdOption.OptionTypes.ipv4_multicast_option:
                self.content = SomeIpSdOptions.SdOption.SdIpv4MulticastOption(self._io, self, self._root)
                self.content._read()
            elif _on == SomeIpSdOptions.SdOption.OptionTypes.ipv6_endpoint_option:
                self.content = SomeIpSdOptions.SdOption.SdIpv6EndpointOption(self._io, self, self._root)
                self.content._read()
            elif _on == SomeIpSdOptions.SdOption.OptionTypes.ipv6_multicast_option:
                self.content = SomeIpSdOptions.SdOption.SdIpv6MulticastOption(self._io, self, self._root)
                self.content._read()
            self._debug['content']['end'] = self._io.pos()

        class SdOptionHeader(KaitaiStruct):
            SEQ_FIELDS = ["length", "type"]
            def __init__(self, _io, _parent=None, _root=None):
                self._io = _io
                self._parent = _parent
                self._root = _root if _root else self
                self._debug = collections.defaultdict(dict)

            def _read(self):
                self._debug['length']['start'] = self._io.pos()
                self.length = self._io.read_u2be()
                self._debug['length']['end'] = self._io.pos()
                self._debug['type']['start'] = self._io.pos()
                self.type = KaitaiStream.resolve_enum(SomeIpSdOptions.SdOption.OptionTypes, self._io.read_u1())
                self._debug['type']['end'] = self._io.pos()


        class SdConfigString(KaitaiStruct):
            SEQ_FIELDS = ["length", "config"]
            def __init__(self, _io, _parent=None, _root=None):
                self._io = _io
                self._parent = _parent
                self._root = _root if _root else self
                self._debug = collections.defaultdict(dict)

            def _read(self):
                self._debug['length']['start'] = self._io.pos()
                self.length = self._io.read_u1()
                self._debug['length']['end'] = self._io.pos()
                if self.length != 0:
                    self._debug['config']['start'] = self._io.pos()
                    self._raw_config = self._io.read_bytes(self.length)
                    _io__raw_config = KaitaiStream(BytesIO(self._raw_config))
                    self.config = SomeIpSdOptions.SdOption.SdConfigKvPair(_io__raw_config, self, self._root)
                    self.config._read()
                    self._debug['config']['end'] = self._io.pos()



        class SdConfigStringsContainer(KaitaiStruct):
            SEQ_FIELDS = ["config_strings"]
            def __init__(self, _io, _parent=None, _root=None):
                self._io = _io
                self._parent = _parent
                self._root = _root if _root else self
                self._debug = collections.defaultdict(dict)

            def _read(self):
                self._debug['config_strings']['start'] = self._io.pos()
                self.config_strings = []
                i = 0
                while not self._io.is_eof():
                    if not 'arr' in self._debug['config_strings']:
                        self._debug['config_strings']['arr'] = []
                    self._debug['config_strings']['arr'].append({'start': self._io.pos()})
                    _t_config_strings = SomeIpSdOptions.SdOption.SdConfigString(self._io, self, self._root)
                    _t_config_strings._read()
                    self.config_strings.append(_t_config_strings)
                    self._debug['config_strings']['arr'][len(self.config_strings) - 1]['end'] = self._io.pos()
                    i += 1

                self._debug['config_strings']['end'] = self._io.pos()


        class SdConfigurationOption(KaitaiStruct):
            SEQ_FIELDS = ["reserved", "configurations"]
            def __init__(self, _io, _parent=None, _root=None):
                self._io = _io
                self._parent = _parent
                self._root = _root if _root else self
                self._debug = collections.defaultdict(dict)

            def _read(self):
                self._debug['reserved']['start'] = self._io.pos()
                self.reserved = self._io.read_u1()
                self._debug['reserved']['end'] = self._io.pos()
                self._debug['configurations']['start'] = self._io.pos()
                self._raw_configurations = self._io.read_bytes((self._parent.header.length - 1))
                _io__raw_configurations = KaitaiStream(BytesIO(self._raw_configurations))
                self.configurations = SomeIpSdOptions.SdOption.SdConfigStringsContainer(_io__raw_configurations, self, self._root)
                self.configurations._read()
                self._debug['configurations']['end'] = self._io.pos()


        class SdIpv4MulticastOption(KaitaiStruct):
            SEQ_FIELDS = ["reserved", "address", "reserved2", "l4_protocol", "port"]
            def __init__(self, _io, _parent=None, _root=None):
                self._io = _io
                self._parent = _parent
                self._root = _root if _root else self
                self._debug = collections.defaultdict(dict)

            def _read(self):
                self._debug['reserved']['start'] = self._io.pos()
                self.reserved = self._io.read_u1()
                self._debug['reserved']['end'] = self._io.pos()
                self._debug['address']['start'] = self._io.pos()
                self.address = self._io.read_bytes(4)
                self._debug['address']['end'] = self._io.pos()
                self._debug['reserved2']['start'] = self._io.pos()
                self.reserved2 = self._io.read_u1()
                self._debug['reserved2']['end'] = self._io.pos()
                self._debug['l4_protocol']['start'] = self._io.pos()
                self.l4_protocol = self._io.read_u1()
                self._debug['l4_protocol']['end'] = self._io.pos()
                self._debug['port']['start'] = self._io.pos()
                self.port = self._io.read_u2be()
                self._debug['port']['end'] = self._io.pos()


        class SdIpv4SdEndpointOption(KaitaiStruct):
            SEQ_FIELDS = ["reserved", "address", "reserved2", "l4_protocol", "port"]
            def __init__(self, _io, _parent=None, _root=None):
                self._io = _io
                self._parent = _parent
                self._root = _root if _root else self
                self._debug = collections.defaultdict(dict)

            def _read(self):
                self._debug['reserved']['start'] = self._io.pos()
                self.reserved = self._io.read_u1()
                self._debug['reserved']['end'] = self._io.pos()
                self._debug['address']['start'] = self._io.pos()
                self.address = self._io.read_bytes(4)
                self._debug['address']['end'] = self._io.pos()
                self._debug['reserved2']['start'] = self._io.pos()
                self.reserved2 = self._io.read_u1()
                self._debug['reserved2']['end'] = self._io.pos()
                self._debug['l4_protocol']['start'] = self._io.pos()
                self.l4_protocol = self._io.read_u1()
                self._debug['l4_protocol']['end'] = self._io.pos()
                self._debug['port']['start'] = self._io.pos()
                self.port = self._io.read_u2be()
                self._debug['port']['end'] = self._io.pos()


        class SdIpv6MulticastOption(KaitaiStruct):
            SEQ_FIELDS = ["reserved", "address", "reserved2", "l4_protocol", "port"]
            def __init__(self, _io, _parent=None, _root=None):
                self._io = _io
                self._parent = _parent
                self._root = _root if _root else self
                self._debug = collections.defaultdict(dict)

            def _read(self):
                self._debug['reserved']['start'] = self._io.pos()
                self.reserved = self._io.read_u1()
                self._debug['reserved']['end'] = self._io.pos()
                self._debug['address']['start'] = self._io.pos()
                self.address = self._io.read_bytes(16)
                self._debug['address']['end'] = self._io.pos()
                self._debug['reserved2']['start'] = self._io.pos()
                self.reserved2 = self._io.read_u1()
                self._debug['reserved2']['end'] = self._io.pos()
                self._debug['l4_protocol']['start'] = self._io.pos()
                self.l4_protocol = self._io.read_u1()
                self._debug['l4_protocol']['end'] = self._io.pos()
                self._debug['port']['start'] = self._io.pos()
                self.port = self._io.read_u2be()
                self._debug['port']['end'] = self._io.pos()


        class SdConfigKvPair(KaitaiStruct):
            SEQ_FIELDS = ["key", "value"]
            def __init__(self, _io, _parent=None, _root=None):
                self._io = _io
                self._parent = _parent
                self._root = _root if _root else self
                self._debug = collections.defaultdict(dict)

            def _read(self):
                self._debug['key']['start'] = self._io.pos()
                self.key = (self._io.read_bytes_term(61, False, True, True)).decode(u"ASCII")
                self._debug['key']['end'] = self._io.pos()
                self._debug['value']['start'] = self._io.pos()
                self.value = (self._io.read_bytes_full()).decode(u"ASCII")
                self._debug['value']['end'] = self._io.pos()


        class SdIpv6SdEndpointOption(KaitaiStruct):
            SEQ_FIELDS = ["reserved", "address", "reserved2", "l4_protocol", "port"]
            def __init__(self, _io, _parent=None, _root=None):
                self._io = _io
                self._parent = _parent
                self._root = _root if _root else self
                self._debug = collections.defaultdict(dict)

            def _read(self):
                self._debug['reserved']['start'] = self._io.pos()
                self.reserved = self._io.read_u1()
                self._debug['reserved']['end'] = self._io.pos()
                self._debug['address']['start'] = self._io.pos()
                self.address = self._io.read_bytes(16)
                self._debug['address']['end'] = self._io.pos()
                self._debug['reserved2']['start'] = self._io.pos()
                self.reserved2 = self._io.read_u1()
                self._debug['reserved2']['end'] = self._io.pos()
                self._debug['l4_protocol']['start'] = self._io.pos()
                self.l4_protocol = self._io.read_u1()
                self._debug['l4_protocol']['end'] = self._io.pos()
                self._debug['port']['start'] = self._io.pos()
                self.port = self._io.read_u2be()
                self._debug['port']['end'] = self._io.pos()


        class SdIpv4EndpointOption(KaitaiStruct):
            SEQ_FIELDS = ["reserved", "address", "reserved2", "l4_protocol", "port"]
            def __init__(self, _io, _parent=None, _root=None):
                self._io = _io
                self._parent = _parent
                self._root = _root if _root else self
                self._debug = collections.defaultdict(dict)

            def _read(self):
                self._debug['reserved']['start'] = self._io.pos()
                self.reserved = self._io.read_u1()
                self._debug['reserved']['end'] = self._io.pos()
                self._debug['address']['start'] = self._io.pos()
                self.address = self._io.read_bytes(4)
                self._debug['address']['end'] = self._io.pos()
                self._debug['reserved2']['start'] = self._io.pos()
                self.reserved2 = self._io.read_u1()
                self._debug['reserved2']['end'] = self._io.pos()
                self._debug['l4_protocol']['start'] = self._io.pos()
                self.l4_protocol = self._io.read_u1()
                self._debug['l4_protocol']['end'] = self._io.pos()
                self._debug['port']['start'] = self._io.pos()
                self.port = self._io.read_u2be()
                self._debug['port']['end'] = self._io.pos()


        class SdIpv6EndpointOption(KaitaiStruct):
            SEQ_FIELDS = ["reserved", "address", "reserved2", "l4_protocol", "port"]
            def __init__(self, _io, _parent=None, _root=None):
                self._io = _io
                self._parent = _parent
                self._root = _root if _root else self
                self._debug = collections.defaultdict(dict)

            def _read(self):
                self._debug['reserved']['start'] = self._io.pos()
                self.reserved = self._io.read_u1()
                self._debug['reserved']['end'] = self._io.pos()
                self._debug['address']['start'] = self._io.pos()
                self.address = self._io.read_bytes(16)
                self._debug['address']['end'] = self._io.pos()
                self._debug['reserved2']['start'] = self._io.pos()
                self.reserved2 = self._io.read_u1()
                self._debug['reserved2']['end'] = self._io.pos()
                self._debug['l4_protocol']['start'] = self._io.pos()
                self.l4_protocol = self._io.read_u1()
                self._debug['l4_protocol']['end'] = self._io.pos()
                self._debug['port']['start'] = self._io.pos()
                self.port = self._io.read_u2be()
                self._debug['port']['end'] = self._io.pos()


        class SdLoadBalancingOption(KaitaiStruct):
            SEQ_FIELDS = ["reserved", "priority", "weight"]
            def __init__(self, _io, _parent=None, _root=None):
                self._io = _io
                self._parent = _parent
                self._root = _root if _root else self
                self._debug = collections.defaultdict(dict)

            def _read(self):
                self._debug['reserved']['start'] = self._io.pos()
                self.reserved = self._io.read_u1()
                self._debug['reserved']['end'] = self._io.pos()
                self._debug['priority']['start'] = self._io.pos()
                self.priority = self._io.read_u2be()
                self._debug['priority']['end'] = self._io.pos()
                self._debug['weight']['start'] = self._io.pos()
                self.weight = self._io.read_u2be()
                self._debug['weight']['end'] = self._io.pos()







# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

from pkg_resources import parse_version
import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO
import collections


if parse_version(kaitaistruct.__version__) < parse_version('0.9'):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

class AvantesRoh60(KaitaiStruct):
    """Avantes USB spectrometers are supplied with a Windows binary which
    generates one ROH and one RCM file when the user clicks "Save
    experiment". In the version of 6.0, the ROH file contains a header
    of 22 four-byte floats, then the spectrum as a float array and a
    footer of 3 floats. The first and last pixel numbers are specified in the
    header and determine the (length+1) of the spectral data. In the tested
    files, the length is (2032-211-1)=1820 pixels, but Kaitai determines this
    automatically anyway.
    
    The wavelength calibration is stored as a polynomial with coefficients
    of 'wlintercept', 'wlx1', ... 'wlx4', the argument of which is the
    (pixel number + 1), as found out by comparing with the original
    Avantes converted data files. There is no intensity calibration saved,
    but it is recommended to do it in your program - the CCD in the spectrometer
    is so uneven that one should prepare exact pixel-to-pixel calibration curves
    to get reasonable spectral results.
    
    The rest of the header floats is not known to the author. Note that the
    newer version of Avantes software has a different format, see also
    <https://www.mathworks.com/matlabcentral/fileexchange/37103-avantes-to-matlab>
    
    The RCM file contains the user-specified comment, so it may be useful
    for automatic conversion of data. You may wish to divide the spectra by
    the integration time before comparing them.
    
    Written and tested by Filip Dominec, 2017-2018
    """
    SEQ_FIELDS = ["unknown1", "wlintercept", "wlx1", "wlx2", "wlx3", "wlx4", "unknown2", "ipixfirst", "ipixlast", "unknown3", "spectrum", "integration_ms", "averaging", "pixel_smoothing"]
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._debug = collections.defaultdict(dict)

    def _read(self):
        self._debug['unknown1']['start'] = self._io.pos()
        self.unknown1 = self._io.read_f4le()
        self._debug['unknown1']['end'] = self._io.pos()
        self._debug['wlintercept']['start'] = self._io.pos()
        self.wlintercept = self._io.read_f4le()
        self._debug['wlintercept']['end'] = self._io.pos()
        self._debug['wlx1']['start'] = self._io.pos()
        self.wlx1 = self._io.read_f4le()
        self._debug['wlx1']['end'] = self._io.pos()
        self._debug['wlx2']['start'] = self._io.pos()
        self.wlx2 = self._io.read_f4le()
        self._debug['wlx2']['end'] = self._io.pos()
        self._debug['wlx3']['start'] = self._io.pos()
        self.wlx3 = self._io.read_f4le()
        self._debug['wlx3']['end'] = self._io.pos()
        self._debug['wlx4']['start'] = self._io.pos()
        self.wlx4 = self._io.read_f4le()
        self._debug['wlx4']['end'] = self._io.pos()
        self._debug['unknown2']['start'] = self._io.pos()
        self.unknown2 = [None] * (9)
        for i in range(9):
            if not 'arr' in self._debug['unknown2']:
                self._debug['unknown2']['arr'] = []
            self._debug['unknown2']['arr'].append({'start': self._io.pos()})
            self.unknown2[i] = self._io.read_f4le()
            self._debug['unknown2']['arr'][i]['end'] = self._io.pos()

        self._debug['unknown2']['end'] = self._io.pos()
        self._debug['ipixfirst']['start'] = self._io.pos()
        self.ipixfirst = self._io.read_f4le()
        self._debug['ipixfirst']['end'] = self._io.pos()
        self._debug['ipixlast']['start'] = self._io.pos()
        self.ipixlast = self._io.read_f4le()
        self._debug['ipixlast']['end'] = self._io.pos()
        self._debug['unknown3']['start'] = self._io.pos()
        self.unknown3 = [None] * (4)
        for i in range(4):
            if not 'arr' in self._debug['unknown3']:
                self._debug['unknown3']['arr'] = []
            self._debug['unknown3']['arr'].append({'start': self._io.pos()})
            self.unknown3[i] = self._io.read_f4le()
            self._debug['unknown3']['arr'][i]['end'] = self._io.pos()

        self._debug['unknown3']['end'] = self._io.pos()
        self._debug['spectrum']['start'] = self._io.pos()
        self.spectrum = [None] * (((int(self.ipixlast) - int(self.ipixfirst)) - 1))
        for i in range(((int(self.ipixlast) - int(self.ipixfirst)) - 1)):
            if not 'arr' in self._debug['spectrum']:
                self._debug['spectrum']['arr'] = []
            self._debug['spectrum']['arr'].append({'start': self._io.pos()})
            self.spectrum[i] = self._io.read_f4le()
            self._debug['spectrum']['arr'][i]['end'] = self._io.pos()

        self._debug['spectrum']['end'] = self._io.pos()
        self._debug['integration_ms']['start'] = self._io.pos()
        self.integration_ms = self._io.read_f4le()
        self._debug['integration_ms']['end'] = self._io.pos()
        self._debug['averaging']['start'] = self._io.pos()
        self.averaging = self._io.read_f4le()
        self._debug['averaging']['end'] = self._io.pos()
        self._debug['pixel_smoothing']['start'] = self._io.pos()
        self.pixel_smoothing = self._io.read_f4le()
        self._debug['pixel_smoothing']['end'] = self._io.pos()





# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

from pkg_resources import parse_version
import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO
import collections


if parse_version(kaitaistruct.__version__) < parse_version('0.9'):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

class SaintsRow2VppPc(KaitaiStruct):
    SEQ_FIELDS = ["magic", "pad1", "num_files", "container_size", "len_offsets", "len_filenames", "len_extensions", "smth5", "smth6", "smth7", "smth8", "smth9"]
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._debug = collections.defaultdict(dict)

    def _read(self):
        self._debug['magic']['start'] = self._io.pos()
        self.magic = self._io.read_bytes(5)
        self._debug['magic']['end'] = self._io.pos()
        if not self.magic == b"\xCE\x0A\x89\x51\x04":
            raise kaitaistruct.ValidationNotEqualError(b"\xCE\x0A\x89\x51\x04", self.magic, self._io, u"/seq/0")
        self._debug['pad1']['start'] = self._io.pos()
        self.pad1 = self._io.read_bytes(335)
        self._debug['pad1']['end'] = self._io.pos()
        self._debug['num_files']['start'] = self._io.pos()
        self.num_files = self._io.read_s4le()
        self._debug['num_files']['end'] = self._io.pos()
        self._debug['container_size']['start'] = self._io.pos()
        self.container_size = self._io.read_s4le()
        self._debug['container_size']['end'] = self._io.pos()
        self._debug['len_offsets']['start'] = self._io.pos()
        self.len_offsets = self._io.read_s4le()
        self._debug['len_offsets']['end'] = self._io.pos()
        self._debug['len_filenames']['start'] = self._io.pos()
        self.len_filenames = self._io.read_s4le()
        self._debug['len_filenames']['end'] = self._io.pos()
        self._debug['len_extensions']['start'] = self._io.pos()
        self.len_extensions = self._io.read_s4le()
        self._debug['len_extensions']['end'] = self._io.pos()
        self._debug['smth5']['start'] = self._io.pos()
        self.smth5 = self._io.read_s4le()
        self._debug['smth5']['end'] = self._io.pos()
        self._debug['smth6']['start'] = self._io.pos()
        self.smth6 = self._io.read_s4le()
        self._debug['smth6']['end'] = self._io.pos()
        self._debug['smth7']['start'] = self._io.pos()
        self.smth7 = self._io.read_s4le()
        self._debug['smth7']['end'] = self._io.pos()
        self._debug['smth8']['start'] = self._io.pos()
        self.smth8 = self._io.read_s4le()
        self._debug['smth8']['end'] = self._io.pos()
        self._debug['smth9']['start'] = self._io.pos()
        self.smth9 = self._io.read_s4le()
        self._debug['smth9']['end'] = self._io.pos()

    class Offsets(KaitaiStruct):
        SEQ_FIELDS = ["entries"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['entries']['start'] = self._io.pos()
            self.entries = []
            i = 0
            while not self._io.is_eof():
                if not 'arr' in self._debug['entries']:
                    self._debug['entries']['arr'] = []
                self._debug['entries']['arr'].append({'start': self._io.pos()})
                _t_entries = SaintsRow2VppPc.Offsets.Offset(self._io, self, self._root)
                _t_entries._read()
                self.entries.append(_t_entries)
                self._debug['entries']['arr'][len(self.entries) - 1]['end'] = self._io.pos()
                i += 1

            self._debug['entries']['end'] = self._io.pos()

        class Offset(KaitaiStruct):
            SEQ_FIELDS = ["name_ofs", "ext_ofs", "smth2", "ofs_body", "len_body", "always_minus_1", "always_zero"]
            def __init__(self, _io, _parent=None, _root=None):
                self._io = _io
                self._parent = _parent
                self._root = _root if _root else self
                self._debug = collections.defaultdict(dict)

            def _read(self):
                self._debug['name_ofs']['start'] = self._io.pos()
                self.name_ofs = self._io.read_u4le()
                self._debug['name_ofs']['end'] = self._io.pos()
                self._debug['ext_ofs']['start'] = self._io.pos()
                self.ext_ofs = self._io.read_u4le()
                self._debug['ext_ofs']['end'] = self._io.pos()
                self._debug['smth2']['start'] = self._io.pos()
                self.smth2 = self._io.read_s4le()
                self._debug['smth2']['end'] = self._io.pos()
                self._debug['ofs_body']['start'] = self._io.pos()
                self.ofs_body = self._io.read_s4le()
                self._debug['ofs_body']['end'] = self._io.pos()
                self._debug['len_body']['start'] = self._io.pos()
                self.len_body = self._io.read_s4le()
                self._debug['len_body']['end'] = self._io.pos()
                self._debug['always_minus_1']['start'] = self._io.pos()
                self.always_minus_1 = self._io.read_s4le()
                self._debug['always_minus_1']['end'] = self._io.pos()
                self._debug['always_zero']['start'] = self._io.pos()
                self.always_zero = self._io.read_s4le()
                self._debug['always_zero']['end'] = self._io.pos()

            @property
            def filename(self):
                if hasattr(self, '_m_filename'):
                    return self._m_filename if hasattr(self, '_m_filename') else None

                io = self._root.filenames._io
                _pos = io.pos()
                io.seek(self.name_ofs)
                self._debug['_m_filename']['start'] = io.pos()
                self._m_filename = (io.read_bytes_term(0, False, True, True)).decode(u"UTF-8")
                self._debug['_m_filename']['end'] = io.pos()
                io.seek(_pos)
                return self._m_filename if hasattr(self, '_m_filename') else None

            @property
            def ext(self):
                if hasattr(self, '_m_ext'):
                    return self._m_ext if hasattr(self, '_m_ext') else None

                io = self._root.extensions._io
                _pos = io.pos()
                io.seek(self.ext_ofs)
                self._debug['_m_ext']['start'] = io.pos()
                self._m_ext = (io.read_bytes_term(0, False, True, True)).decode(u"UTF-8")
                self._debug['_m_ext']['end'] = io.pos()
                io.seek(_pos)
                return self._m_ext if hasattr(self, '_m_ext') else None

            @property
            def body(self):
                if hasattr(self, '_m_body'):
                    return self._m_body if hasattr(self, '_m_body') else None

                io = self._root._io
                _pos = io.pos()
                io.seek((self._root.data_start + self.ofs_body))
                self._debug['_m_body']['start'] = io.pos()
                self._m_body = io.read_bytes(self.len_body)
                self._debug['_m_body']['end'] = io.pos()
                io.seek(_pos)
                return self._m_body if hasattr(self, '_m_body') else None



    class Strings(KaitaiStruct):
        SEQ_FIELDS = ["entries"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['entries']['start'] = self._io.pos()
            self.entries = []
            i = 0
            while not self._io.is_eof():
                if not 'arr' in self._debug['entries']:
                    self._debug['entries']['arr'] = []
                self._debug['entries']['arr'].append({'start': self._io.pos()})
                self.entries.append((self._io.read_bytes_term(0, False, True, True)).decode(u"UTF-8"))
                self._debug['entries']['arr'][len(self.entries) - 1]['end'] = self._io.pos()
                i += 1

            self._debug['entries']['end'] = self._io.pos()


    @property
    def filenames(self):
        if hasattr(self, '_m_filenames'):
            return self._m_filenames if hasattr(self, '_m_filenames') else None

        _pos = self._io.pos()
        self._io.seek(self.ofs_filenames)
        self._debug['_m_filenames']['start'] = self._io.pos()
        self._raw__m_filenames = self._io.read_bytes(self.len_filenames)
        _io__raw__m_filenames = KaitaiStream(BytesIO(self._raw__m_filenames))
        self._m_filenames = SaintsRow2VppPc.Strings(_io__raw__m_filenames, self, self._root)
        self._m_filenames._read()
        self._debug['_m_filenames']['end'] = self._io.pos()
        self._io.seek(_pos)
        return self._m_filenames if hasattr(self, '_m_filenames') else None

    @property
    def ofs_extensions(self):
        if hasattr(self, '_m_ofs_extensions'):
            return self._m_ofs_extensions if hasattr(self, '_m_ofs_extensions') else None

        self._m_ofs_extensions = (((self.ofs_filenames + self.len_filenames) & 4294965248) + 2048)
        return self._m_ofs_extensions if hasattr(self, '_m_ofs_extensions') else None

    @property
    def files(self):
        if hasattr(self, '_m_files'):
            return self._m_files if hasattr(self, '_m_files') else None

        _pos = self._io.pos()
        self._io.seek(2048)
        self._debug['_m_files']['start'] = self._io.pos()
        self._raw__m_files = self._io.read_bytes(self.len_offsets)
        _io__raw__m_files = KaitaiStream(BytesIO(self._raw__m_files))
        self._m_files = SaintsRow2VppPc.Offsets(_io__raw__m_files, self, self._root)
        self._m_files._read()
        self._debug['_m_files']['end'] = self._io.pos()
        self._io.seek(_pos)
        return self._m_files if hasattr(self, '_m_files') else None

    @property
    def data_start(self):
        if hasattr(self, '_m_data_start'):
            return self._m_data_start if hasattr(self, '_m_data_start') else None

        self._m_data_start = (((self.ofs_extensions + self.len_extensions) & 4294965248) + 2048)
        return self._m_data_start if hasattr(self, '_m_data_start') else None

    @property
    def extensions(self):
        if hasattr(self, '_m_extensions'):
            return self._m_extensions if hasattr(self, '_m_extensions') else None

        _pos = self._io.pos()
        self._io.seek(self.ofs_extensions)
        self._debug['_m_extensions']['start'] = self._io.pos()
        self._raw__m_extensions = self._io.read_bytes(self.len_extensions)
        _io__raw__m_extensions = KaitaiStream(BytesIO(self._raw__m_extensions))
        self._m_extensions = SaintsRow2VppPc.Strings(_io__raw__m_extensions, self, self._root)
        self._m_extensions._read()
        self._debug['_m_extensions']['end'] = self._io.pos()
        self._io.seek(_pos)
        return self._m_extensions if hasattr(self, '_m_extensions') else None

    @property
    def ofs_filenames(self):
        if hasattr(self, '_m_ofs_filenames'):
            return self._m_ofs_filenames if hasattr(self, '_m_ofs_filenames') else None

        self._m_ofs_filenames = (((2048 + self.len_offsets) & 4294965248) + 2048)
        return self._m_ofs_filenames if hasattr(self, '_m_ofs_filenames') else None





# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

from pkg_resources import parse_version
import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO
from enum import Enum
import collections


if parse_version(kaitaistruct.__version__) < parse_version('0.9'):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

from polyfile.kaitai.parsers import ethernet_frame
class PacketPpi(KaitaiStruct):
    """PPI is a standard for link layer packet encapsulation, proposed as
    generic extensible container to store both captured in-band data and
    out-of-band data. Originally it was developed to provide 802.11n
    radio information, but can be used for other purposes as well.
    
    Sample capture: https://wiki.wireshark.org/SampleCaptures?action=AttachFile&do=get&target=Http.cap
    
    .. seealso::
       PPI header format spec, section 3 - https://web.archive.org/web/20090206112419/https://www.cacetech.com/documents/PPI_Header_format_1.0.1.pdf
    """

    class PfhType(Enum):
        radio_802_11_common = 2
        radio_802_11n_mac_ext = 3
        radio_802_11n_mac_phy_ext = 4
        spectrum_map = 5
        process_info = 6
        capture_info = 7

    class Linktype(Enum):
        null_linktype = 0
        ethernet = 1
        ax25 = 3
        ieee802_5 = 6
        arcnet_bsd = 7
        slip = 8
        ppp = 9
        fddi = 10
        ppp_hdlc = 50
        ppp_ether = 51
        atm_rfc1483 = 100
        raw = 101
        c_hdlc = 104
        ieee802_11 = 105
        frelay = 107
        loop = 108
        linux_sll = 113
        ltalk = 114
        pflog = 117
        ieee802_11_prism = 119
        ip_over_fc = 122
        sunatm = 123
        ieee802_11_radiotap = 127
        arcnet_linux = 129
        apple_ip_over_ieee1394 = 138
        mtp2_with_phdr = 139
        mtp2 = 140
        mtp3 = 141
        sccp = 142
        docsis = 143
        linux_irda = 144
        user0 = 147
        user1 = 148
        user2 = 149
        user3 = 150
        user4 = 151
        user5 = 152
        user6 = 153
        user7 = 154
        user8 = 155
        user9 = 156
        user10 = 157
        user11 = 158
        user12 = 159
        user13 = 160
        user14 = 161
        user15 = 162
        ieee802_11_avs = 163
        bacnet_ms_tp = 165
        ppp_pppd = 166
        gprs_llc = 169
        gpf_t = 170
        gpf_f = 171
        linux_lapd = 177
        bluetooth_hci_h4 = 187
        usb_linux = 189
        ppi = 192
        ieee802_15_4 = 195
        sita = 196
        erf = 197
        bluetooth_hci_h4_with_phdr = 201
        ax25_kiss = 202
        lapd = 203
        ppp_with_dir = 204
        c_hdlc_with_dir = 205
        frelay_with_dir = 206
        ipmb_linux = 209
        ieee802_15_4_nonask_phy = 215
        usb_linux_mmapped = 220
        fc_2 = 224
        fc_2_with_frame_delims = 225
        ipnet = 226
        can_socketcan = 227
        ipv4 = 228
        ipv6 = 229
        ieee802_15_4_nofcs = 230
        dbus = 231
        dvb_ci = 235
        mux27010 = 236
        stanag_5066_d_pdu = 237
        nflog = 239
        netanalyzer = 240
        netanalyzer_transparent = 241
        ipoib = 242
        mpeg_2_ts = 243
        ng40 = 244
        nfc_llcp = 245
        infiniband = 247
        sctp = 248
        usbpcap = 249
        rtac_serial = 250
        bluetooth_le_ll = 251
        netlink = 253
        bluetooth_linux_monitor = 254
        bluetooth_bredr_bb = 255
        bluetooth_le_ll_with_phdr = 256
        profibus_dl = 257
        pktap = 258
        epon = 259
        ipmi_hpm_2 = 260
        zwave_r1_r2 = 261
        zwave_r3 = 262
        wattstopper_dlm = 263
        iso_14443 = 264
    SEQ_FIELDS = ["header", "fields", "body"]
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._debug = collections.defaultdict(dict)

    def _read(self):
        self._debug['header']['start'] = self._io.pos()
        self.header = PacketPpi.PacketPpiHeader(self._io, self, self._root)
        self.header._read()
        self._debug['header']['end'] = self._io.pos()
        self._debug['fields']['start'] = self._io.pos()
        self._raw_fields = self._io.read_bytes((self.header.pph_len - 8))
        _io__raw_fields = KaitaiStream(BytesIO(self._raw_fields))
        self.fields = PacketPpi.PacketPpiFields(_io__raw_fields, self, self._root)
        self.fields._read()
        self._debug['fields']['end'] = self._io.pos()
        self._debug['body']['start'] = self._io.pos()
        _on = self.header.pph_dlt
        if _on == PacketPpi.Linktype.ppi:
            self._raw_body = self._io.read_bytes_full()
            _io__raw_body = KaitaiStream(BytesIO(self._raw_body))
            self.body = PacketPpi(_io__raw_body)
            self.body._read()
        elif _on == PacketPpi.Linktype.ethernet:
            self._raw_body = self._io.read_bytes_full()
            _io__raw_body = KaitaiStream(BytesIO(self._raw_body))
            self.body = ethernet_frame.EthernetFrame(_io__raw_body)
            self.body._read()
        else:
            self.body = self._io.read_bytes_full()
        self._debug['body']['end'] = self._io.pos()

    class PacketPpiFields(KaitaiStruct):
        SEQ_FIELDS = ["entries"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['entries']['start'] = self._io.pos()
            self.entries = []
            i = 0
            while not self._io.is_eof():
                if not 'arr' in self._debug['entries']:
                    self._debug['entries']['arr'] = []
                self._debug['entries']['arr'].append({'start': self._io.pos()})
                _t_entries = PacketPpi.PacketPpiField(self._io, self, self._root)
                _t_entries._read()
                self.entries.append(_t_entries)
                self._debug['entries']['arr'][len(self.entries) - 1]['end'] = self._io.pos()
                i += 1

            self._debug['entries']['end'] = self._io.pos()


    class Radio80211nMacExtBody(KaitaiStruct):
        """
        .. seealso::
           PPI header format spec, section 4.1.3 - https://web.archive.org/web/20090206112419/https://www.cacetech.com/documents/PPI_Header_format_1.0.1.pdf
        """
        SEQ_FIELDS = ["flags", "a_mpdu_id", "num_delimiters", "reserved"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['flags']['start'] = self._io.pos()
            self.flags = PacketPpi.MacFlags(self._io, self, self._root)
            self.flags._read()
            self._debug['flags']['end'] = self._io.pos()
            self._debug['a_mpdu_id']['start'] = self._io.pos()
            self.a_mpdu_id = self._io.read_u4le()
            self._debug['a_mpdu_id']['end'] = self._io.pos()
            self._debug['num_delimiters']['start'] = self._io.pos()
            self.num_delimiters = self._io.read_u1()
            self._debug['num_delimiters']['end'] = self._io.pos()
            self._debug['reserved']['start'] = self._io.pos()
            self.reserved = self._io.read_bytes(3)
            self._debug['reserved']['end'] = self._io.pos()


    class MacFlags(KaitaiStruct):
        SEQ_FIELDS = ["unused1", "aggregate_delimiter", "more_aggregates", "aggregate", "dup_rx", "rx_short_guard", "is_ht_40", "greenfield", "unused2"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['unused1']['start'] = self._io.pos()
            self.unused1 = self._io.read_bits_int_be(1) != 0
            self._debug['unused1']['end'] = self._io.pos()
            self._debug['aggregate_delimiter']['start'] = self._io.pos()
            self.aggregate_delimiter = self._io.read_bits_int_be(1) != 0
            self._debug['aggregate_delimiter']['end'] = self._io.pos()
            self._debug['more_aggregates']['start'] = self._io.pos()
            self.more_aggregates = self._io.read_bits_int_be(1) != 0
            self._debug['more_aggregates']['end'] = self._io.pos()
            self._debug['aggregate']['start'] = self._io.pos()
            self.aggregate = self._io.read_bits_int_be(1) != 0
            self._debug['aggregate']['end'] = self._io.pos()
            self._debug['dup_rx']['start'] = self._io.pos()
            self.dup_rx = self._io.read_bits_int_be(1) != 0
            self._debug['dup_rx']['end'] = self._io.pos()
            self._debug['rx_short_guard']['start'] = self._io.pos()
            self.rx_short_guard = self._io.read_bits_int_be(1) != 0
            self._debug['rx_short_guard']['end'] = self._io.pos()
            self._debug['is_ht_40']['start'] = self._io.pos()
            self.is_ht_40 = self._io.read_bits_int_be(1) != 0
            self._debug['is_ht_40']['end'] = self._io.pos()
            self._debug['greenfield']['start'] = self._io.pos()
            self.greenfield = self._io.read_bits_int_be(1) != 0
            self._debug['greenfield']['end'] = self._io.pos()
            self._io.align_to_byte()
            self._debug['unused2']['start'] = self._io.pos()
            self.unused2 = self._io.read_bytes(3)
            self._debug['unused2']['end'] = self._io.pos()


    class PacketPpiHeader(KaitaiStruct):
        """
        .. seealso::
           PPI header format spec, section 3.1 - https://web.archive.org/web/20090206112419/https://www.cacetech.com/documents/PPI_Header_format_1.0.1.pdf
        """
        SEQ_FIELDS = ["pph_version", "pph_flags", "pph_len", "pph_dlt"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['pph_version']['start'] = self._io.pos()
            self.pph_version = self._io.read_u1()
            self._debug['pph_version']['end'] = self._io.pos()
            self._debug['pph_flags']['start'] = self._io.pos()
            self.pph_flags = self._io.read_u1()
            self._debug['pph_flags']['end'] = self._io.pos()
            self._debug['pph_len']['start'] = self._io.pos()
            self.pph_len = self._io.read_u2le()
            self._debug['pph_len']['end'] = self._io.pos()
            self._debug['pph_dlt']['start'] = self._io.pos()
            self.pph_dlt = KaitaiStream.resolve_enum(PacketPpi.Linktype, self._io.read_u4le())
            self._debug['pph_dlt']['end'] = self._io.pos()


    class Radio80211CommonBody(KaitaiStruct):
        """
        .. seealso::
           PPI header format spec, section 4.1.2 - https://web.archive.org/web/20090206112419/https://www.cacetech.com/documents/PPI_Header_format_1.0.1.pdf
        """
        SEQ_FIELDS = ["tsf_timer", "flags", "rate", "channel_freq", "channel_flags", "fhss_hopset", "fhss_pattern", "dbm_antsignal", "dbm_antnoise"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['tsf_timer']['start'] = self._io.pos()
            self.tsf_timer = self._io.read_u8le()
            self._debug['tsf_timer']['end'] = self._io.pos()
            self._debug['flags']['start'] = self._io.pos()
            self.flags = self._io.read_u2le()
            self._debug['flags']['end'] = self._io.pos()
            self._debug['rate']['start'] = self._io.pos()
            self.rate = self._io.read_u2le()
            self._debug['rate']['end'] = self._io.pos()
            self._debug['channel_freq']['start'] = self._io.pos()
            self.channel_freq = self._io.read_u2le()
            self._debug['channel_freq']['end'] = self._io.pos()
            self._debug['channel_flags']['start'] = self._io.pos()
            self.channel_flags = self._io.read_u2le()
            self._debug['channel_flags']['end'] = self._io.pos()
            self._debug['fhss_hopset']['start'] = self._io.pos()
            self.fhss_hopset = self._io.read_u1()
            self._debug['fhss_hopset']['end'] = self._io.pos()
            self._debug['fhss_pattern']['start'] = self._io.pos()
            self.fhss_pattern = self._io.read_u1()
            self._debug['fhss_pattern']['end'] = self._io.pos()
            self._debug['dbm_antsignal']['start'] = self._io.pos()
            self.dbm_antsignal = self._io.read_s1()
            self._debug['dbm_antsignal']['end'] = self._io.pos()
            self._debug['dbm_antnoise']['start'] = self._io.pos()
            self.dbm_antnoise = self._io.read_s1()
            self._debug['dbm_antnoise']['end'] = self._io.pos()


    class PacketPpiField(KaitaiStruct):
        """
        .. seealso::
           PPI header format spec, section 3.1 - https://web.archive.org/web/20090206112419/https://www.cacetech.com/documents/PPI_Header_format_1.0.1.pdf
        """
        SEQ_FIELDS = ["pfh_type", "pfh_datalen", "body"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['pfh_type']['start'] = self._io.pos()
            self.pfh_type = KaitaiStream.resolve_enum(PacketPpi.PfhType, self._io.read_u2le())
            self._debug['pfh_type']['end'] = self._io.pos()
            self._debug['pfh_datalen']['start'] = self._io.pos()
            self.pfh_datalen = self._io.read_u2le()
            self._debug['pfh_datalen']['end'] = self._io.pos()
            self._debug['body']['start'] = self._io.pos()
            _on = self.pfh_type
            if _on == PacketPpi.PfhType.radio_802_11_common:
                self._raw_body = self._io.read_bytes(self.pfh_datalen)
                _io__raw_body = KaitaiStream(BytesIO(self._raw_body))
                self.body = PacketPpi.Radio80211CommonBody(_io__raw_body, self, self._root)
                self.body._read()
            elif _on == PacketPpi.PfhType.radio_802_11n_mac_ext:
                self._raw_body = self._io.read_bytes(self.pfh_datalen)
                _io__raw_body = KaitaiStream(BytesIO(self._raw_body))
                self.body = PacketPpi.Radio80211nMacExtBody(_io__raw_body, self, self._root)
                self.body._read()
            elif _on == PacketPpi.PfhType.radio_802_11n_mac_phy_ext:
                self._raw_body = self._io.read_bytes(self.pfh_datalen)
                _io__raw_body = KaitaiStream(BytesIO(self._raw_body))
                self.body = PacketPpi.Radio80211nMacPhyExtBody(_io__raw_body, self, self._root)
                self.body._read()
            else:
                self.body = self._io.read_bytes(self.pfh_datalen)
            self._debug['body']['end'] = self._io.pos()


    class Radio80211nMacPhyExtBody(KaitaiStruct):
        """
        .. seealso::
           PPI header format spec, section 4.1.4 - https://web.archive.org/web/20090206112419/https://www.cacetech.com/documents/PPI_Header_format_1.0.1.pdf
        """
        SEQ_FIELDS = ["flags", "a_mpdu_id", "num_delimiters", "mcs", "num_streams", "rssi_combined", "rssi_ant_ctl", "rssi_ant_ext", "ext_channel_freq", "ext_channel_flags", "rf_signal_noise", "evm"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['flags']['start'] = self._io.pos()
            self.flags = PacketPpi.MacFlags(self._io, self, self._root)
            self.flags._read()
            self._debug['flags']['end'] = self._io.pos()
            self._debug['a_mpdu_id']['start'] = self._io.pos()
            self.a_mpdu_id = self._io.read_u4le()
            self._debug['a_mpdu_id']['end'] = self._io.pos()
            self._debug['num_delimiters']['start'] = self._io.pos()
            self.num_delimiters = self._io.read_u1()
            self._debug['num_delimiters']['end'] = self._io.pos()
            self._debug['mcs']['start'] = self._io.pos()
            self.mcs = self._io.read_u1()
            self._debug['mcs']['end'] = self._io.pos()
            self._debug['num_streams']['start'] = self._io.pos()
            self.num_streams = self._io.read_u1()
            self._debug['num_streams']['end'] = self._io.pos()
            self._debug['rssi_combined']['start'] = self._io.pos()
            self.rssi_combined = self._io.read_u1()
            self._debug['rssi_combined']['end'] = self._io.pos()
            self._debug['rssi_ant_ctl']['start'] = self._io.pos()
            self.rssi_ant_ctl = [None] * (4)
            for i in range(4):
                if not 'arr' in self._debug['rssi_ant_ctl']:
                    self._debug['rssi_ant_ctl']['arr'] = []
                self._debug['rssi_ant_ctl']['arr'].append({'start': self._io.pos()})
                self.rssi_ant_ctl[i] = self._io.read_u1()
                self._debug['rssi_ant_ctl']['arr'][i]['end'] = self._io.pos()

            self._debug['rssi_ant_ctl']['end'] = self._io.pos()
            self._debug['rssi_ant_ext']['start'] = self._io.pos()
            self.rssi_ant_ext = [None] * (4)
            for i in range(4):
                if not 'arr' in self._debug['rssi_ant_ext']:
                    self._debug['rssi_ant_ext']['arr'] = []
                self._debug['rssi_ant_ext']['arr'].append({'start': self._io.pos()})
                self.rssi_ant_ext[i] = self._io.read_u1()
                self._debug['rssi_ant_ext']['arr'][i]['end'] = self._io.pos()

            self._debug['rssi_ant_ext']['end'] = self._io.pos()
            self._debug['ext_channel_freq']['start'] = self._io.pos()
            self.ext_channel_freq = self._io.read_u2le()
            self._debug['ext_channel_freq']['end'] = self._io.pos()
            self._debug['ext_channel_flags']['start'] = self._io.pos()
            self.ext_channel_flags = PacketPpi.Radio80211nMacPhyExtBody.ChannelFlags(self._io, self, self._root)
            self.ext_channel_flags._read()
            self._debug['ext_channel_flags']['end'] = self._io.pos()
            self._debug['rf_signal_noise']['start'] = self._io.pos()
            self.rf_signal_noise = [None] * (4)
            for i in range(4):
                if not 'arr' in self._debug['rf_signal_noise']:
                    self._debug['rf_signal_noise']['arr'] = []
                self._debug['rf_signal_noise']['arr'].append({'start': self._io.pos()})
                _t_rf_signal_noise = PacketPpi.Radio80211nMacPhyExtBody.SignalNoise(self._io, self, self._root)
                _t_rf_signal_noise._read()
                self.rf_signal_noise[i] = _t_rf_signal_noise
                self._debug['rf_signal_noise']['arr'][i]['end'] = self._io.pos()

            self._debug['rf_signal_noise']['end'] = self._io.pos()
            self._debug['evm']['start'] = self._io.pos()
            self.evm = [None] * (4)
            for i in range(4):
                if not 'arr' in self._debug['evm']:
                    self._debug['evm']['arr'] = []
                self._debug['evm']['arr'].append({'start': self._io.pos()})
                self.evm[i] = self._io.read_u4le()
                self._debug['evm']['arr'][i]['end'] = self._io.pos()

            self._debug['evm']['end'] = self._io.pos()

        class ChannelFlags(KaitaiStruct):
            SEQ_FIELDS = ["spectrum_2ghz", "ofdm", "cck", "turbo", "unused", "gfsk", "dyn_cck_ofdm", "only_passive_scan", "spectrum_5ghz"]
            def __init__(self, _io, _parent=None, _root=None):
                self._io = _io
                self._parent = _parent
                self._root = _root if _root else self
                self._debug = collections.defaultdict(dict)

            def _read(self):
                self._debug['spectrum_2ghz']['start'] = self._io.pos()
                self.spectrum_2ghz = self._io.read_bits_int_be(1) != 0
                self._debug['spectrum_2ghz']['end'] = self._io.pos()
                self._debug['ofdm']['start'] = self._io.pos()
                self.ofdm = self._io.read_bits_int_be(1) != 0
                self._debug['ofdm']['end'] = self._io.pos()
                self._debug['cck']['start'] = self._io.pos()
                self.cck = self._io.read_bits_int_be(1) != 0
                self._debug['cck']['end'] = self._io.pos()
                self._debug['turbo']['start'] = self._io.pos()
                self.turbo = self._io.read_bits_int_be(1) != 0
                self._debug['turbo']['end'] = self._io.pos()
                self._debug['unused']['start'] = self._io.pos()
                self.unused = self._io.read_bits_int_be(8)
                self._debug['unused']['end'] = self._io.pos()
                self._debug['gfsk']['start'] = self._io.pos()
                self.gfsk = self._io.read_bits_int_be(1) != 0
                self._debug['gfsk']['end'] = self._io.pos()
                self._debug['dyn_cck_ofdm']['start'] = self._io.pos()
                self.dyn_cck_ofdm = self._io.read_bits_int_be(1) != 0
                self._debug['dyn_cck_ofdm']['end'] = self._io.pos()
                self._debug['only_passive_scan']['start'] = self._io.pos()
                self.only_passive_scan = self._io.read_bits_int_be(1) != 0
                self._debug['only_passive_scan']['end'] = self._io.pos()
                self._debug['spectrum_5ghz']['start'] = self._io.pos()
                self.spectrum_5ghz = self._io.read_bits_int_be(1) != 0
                self._debug['spectrum_5ghz']['end'] = self._io.pos()


        class SignalNoise(KaitaiStruct):
            """RF signal + noise pair at a single antenna."""
            SEQ_FIELDS = ["signal", "noise"]
            def __init__(self, _io, _parent=None, _root=None):
                self._io = _io
                self._parent = _parent
                self._root = _root if _root else self
                self._debug = collections.defaultdict(dict)

            def _read(self):
                self._debug['signal']['start'] = self._io.pos()
                self.signal = self._io.read_s1()
                self._debug['signal']['end'] = self._io.pos()
                self._debug['noise']['start'] = self._io.pos()
                self.noise = self._io.read_s1()
                self._debug['noise']['end'] = self._io.pos()







# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

from pkg_resources import parse_version
import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO
import collections
from enum import Enum


if parse_version(kaitaistruct.__version__) < parse_version('0.9'):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

class Ext2(KaitaiStruct):
    SEQ_FIELDS = []
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._debug = collections.defaultdict(dict)

    def _read(self):
        pass

    class SuperBlockStruct(KaitaiStruct):

        class StateEnum(Enum):
            valid_fs = 1
            error_fs = 2

        class ErrorsEnum(Enum):
            act_continue = 1
            act_ro = 2
            act_panic = 3
        SEQ_FIELDS = ["inodes_count", "blocks_count", "r_blocks_count", "free_blocks_count", "free_inodes_count", "first_data_block", "log_block_size", "log_frag_size", "blocks_per_group", "frags_per_group", "inodes_per_group", "mtime", "wtime", "mnt_count", "max_mnt_count", "magic", "state", "errors", "minor_rev_level", "lastcheck", "checkinterval", "creator_os", "rev_level", "def_resuid", "def_resgid", "first_ino", "inode_size", "block_group_nr", "feature_compat", "feature_incompat", "feature_ro_compat", "uuid", "volume_name", "last_mounted", "algo_bitmap", "prealloc_blocks", "prealloc_dir_blocks", "padding1", "journal_uuid", "journal_inum", "journal_dev", "last_orphan", "hash_seed", "def_hash_version"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['inodes_count']['start'] = self._io.pos()
            self.inodes_count = self._io.read_u4le()
            self._debug['inodes_count']['end'] = self._io.pos()
            self._debug['blocks_count']['start'] = self._io.pos()
            self.blocks_count = self._io.read_u4le()
            self._debug['blocks_count']['end'] = self._io.pos()
            self._debug['r_blocks_count']['start'] = self._io.pos()
            self.r_blocks_count = self._io.read_u4le()
            self._debug['r_blocks_count']['end'] = self._io.pos()
            self._debug['free_blocks_count']['start'] = self._io.pos()
            self.free_blocks_count = self._io.read_u4le()
            self._debug['free_blocks_count']['end'] = self._io.pos()
            self._debug['free_inodes_count']['start'] = self._io.pos()
            self.free_inodes_count = self._io.read_u4le()
            self._debug['free_inodes_count']['end'] = self._io.pos()
            self._debug['first_data_block']['start'] = self._io.pos()
            self.first_data_block = self._io.read_u4le()
            self._debug['first_data_block']['end'] = self._io.pos()
            self._debug['log_block_size']['start'] = self._io.pos()
            self.log_block_size = self._io.read_u4le()
            self._debug['log_block_size']['end'] = self._io.pos()
            self._debug['log_frag_size']['start'] = self._io.pos()
            self.log_frag_size = self._io.read_u4le()
            self._debug['log_frag_size']['end'] = self._io.pos()
            self._debug['blocks_per_group']['start'] = self._io.pos()
            self.blocks_per_group = self._io.read_u4le()
            self._debug['blocks_per_group']['end'] = self._io.pos()
            self._debug['frags_per_group']['start'] = self._io.pos()
            self.frags_per_group = self._io.read_u4le()
            self._debug['frags_per_group']['end'] = self._io.pos()
            self._debug['inodes_per_group']['start'] = self._io.pos()
            self.inodes_per_group = self._io.read_u4le()
            self._debug['inodes_per_group']['end'] = self._io.pos()
            self._debug['mtime']['start'] = self._io.pos()
            self.mtime = self._io.read_u4le()
            self._debug['mtime']['end'] = self._io.pos()
            self._debug['wtime']['start'] = self._io.pos()
            self.wtime = self._io.read_u4le()
            self._debug['wtime']['end'] = self._io.pos()
            self._debug['mnt_count']['start'] = self._io.pos()
            self.mnt_count = self._io.read_u2le()
            self._debug['mnt_count']['end'] = self._io.pos()
            self._debug['max_mnt_count']['start'] = self._io.pos()
            self.max_mnt_count = self._io.read_u2le()
            self._debug['max_mnt_count']['end'] = self._io.pos()
            self._debug['magic']['start'] = self._io.pos()
            self.magic = self._io.read_bytes(2)
            self._debug['magic']['end'] = self._io.pos()
            if not self.magic == b"\x53\xEF":
                raise kaitaistruct.ValidationNotEqualError(b"\x53\xEF", self.magic, self._io, u"/types/super_block_struct/seq/15")
            self._debug['state']['start'] = self._io.pos()
            self.state = KaitaiStream.resolve_enum(Ext2.SuperBlockStruct.StateEnum, self._io.read_u2le())
            self._debug['state']['end'] = self._io.pos()
            self._debug['errors']['start'] = self._io.pos()
            self.errors = KaitaiStream.resolve_enum(Ext2.SuperBlockStruct.ErrorsEnum, self._io.read_u2le())
            self._debug['errors']['end'] = self._io.pos()
            self._debug['minor_rev_level']['start'] = self._io.pos()
            self.minor_rev_level = self._io.read_u2le()
            self._debug['minor_rev_level']['end'] = self._io.pos()
            self._debug['lastcheck']['start'] = self._io.pos()
            self.lastcheck = self._io.read_u4le()
            self._debug['lastcheck']['end'] = self._io.pos()
            self._debug['checkinterval']['start'] = self._io.pos()
            self.checkinterval = self._io.read_u4le()
            self._debug['checkinterval']['end'] = self._io.pos()
            self._debug['creator_os']['start'] = self._io.pos()
            self.creator_os = self._io.read_u4le()
            self._debug['creator_os']['end'] = self._io.pos()
            self._debug['rev_level']['start'] = self._io.pos()
            self.rev_level = self._io.read_u4le()
            self._debug['rev_level']['end'] = self._io.pos()
            self._debug['def_resuid']['start'] = self._io.pos()
            self.def_resuid = self._io.read_u2le()
            self._debug['def_resuid']['end'] = self._io.pos()
            self._debug['def_resgid']['start'] = self._io.pos()
            self.def_resgid = self._io.read_u2le()
            self._debug['def_resgid']['end'] = self._io.pos()
            self._debug['first_ino']['start'] = self._io.pos()
            self.first_ino = self._io.read_u4le()
            self._debug['first_ino']['end'] = self._io.pos()
            self._debug['inode_size']['start'] = self._io.pos()
            self.inode_size = self._io.read_u2le()
            self._debug['inode_size']['end'] = self._io.pos()
            self._debug['block_group_nr']['start'] = self._io.pos()
            self.block_group_nr = self._io.read_u2le()
            self._debug['block_group_nr']['end'] = self._io.pos()
            self._debug['feature_compat']['start'] = self._io.pos()
            self.feature_compat = self._io.read_u4le()
            self._debug['feature_compat']['end'] = self._io.pos()
            self._debug['feature_incompat']['start'] = self._io.pos()
            self.feature_incompat = self._io.read_u4le()
            self._debug['feature_incompat']['end'] = self._io.pos()
            self._debug['feature_ro_compat']['start'] = self._io.pos()
            self.feature_ro_compat = self._io.read_u4le()
            self._debug['feature_ro_compat']['end'] = self._io.pos()
            self._debug['uuid']['start'] = self._io.pos()
            self.uuid = self._io.read_bytes(16)
            self._debug['uuid']['end'] = self._io.pos()
            self._debug['volume_name']['start'] = self._io.pos()
            self.volume_name = self._io.read_bytes(16)
            self._debug['volume_name']['end'] = self._io.pos()
            self._debug['last_mounted']['start'] = self._io.pos()
            self.last_mounted = self._io.read_bytes(64)
            self._debug['last_mounted']['end'] = self._io.pos()
            self._debug['algo_bitmap']['start'] = self._io.pos()
            self.algo_bitmap = self._io.read_u4le()
            self._debug['algo_bitmap']['end'] = self._io.pos()
            self._debug['prealloc_blocks']['start'] = self._io.pos()
            self.prealloc_blocks = self._io.read_u1()
            self._debug['prealloc_blocks']['end'] = self._io.pos()
            self._debug['prealloc_dir_blocks']['start'] = self._io.pos()
            self.prealloc_dir_blocks = self._io.read_u1()
            self._debug['prealloc_dir_blocks']['end'] = self._io.pos()
            self._debug['padding1']['start'] = self._io.pos()
            self.padding1 = self._io.read_bytes(2)
            self._debug['padding1']['end'] = self._io.pos()
            self._debug['journal_uuid']['start'] = self._io.pos()
            self.journal_uuid = self._io.read_bytes(16)
            self._debug['journal_uuid']['end'] = self._io.pos()
            self._debug['journal_inum']['start'] = self._io.pos()
            self.journal_inum = self._io.read_u4le()
            self._debug['journal_inum']['end'] = self._io.pos()
            self._debug['journal_dev']['start'] = self._io.pos()
            self.journal_dev = self._io.read_u4le()
            self._debug['journal_dev']['end'] = self._io.pos()
            self._debug['last_orphan']['start'] = self._io.pos()
            self.last_orphan = self._io.read_u4le()
            self._debug['last_orphan']['end'] = self._io.pos()
            self._debug['hash_seed']['start'] = self._io.pos()
            self.hash_seed = [None] * (4)
            for i in range(4):
                if not 'arr' in self._debug['hash_seed']:
                    self._debug['hash_seed']['arr'] = []
                self._debug['hash_seed']['arr'].append({'start': self._io.pos()})
                self.hash_seed[i] = self._io.read_u4le()
                self._debug['hash_seed']['arr'][i]['end'] = self._io.pos()

            self._debug['hash_seed']['end'] = self._io.pos()
            self._debug['def_hash_version']['start'] = self._io.pos()
            self.def_hash_version = self._io.read_u1()
            self._debug['def_hash_version']['end'] = self._io.pos()

        @property
        def block_size(self):
            if hasattr(self, '_m_block_size'):
                return self._m_block_size if hasattr(self, '_m_block_size') else None

            self._m_block_size = (1024 << self.log_block_size)
            return self._m_block_size if hasattr(self, '_m_block_size') else None

        @property
        def block_group_count(self):
            if hasattr(self, '_m_block_group_count'):
                return self._m_block_group_count if hasattr(self, '_m_block_group_count') else None

            self._m_block_group_count = self.blocks_count // self.blocks_per_group
            return self._m_block_group_count if hasattr(self, '_m_block_group_count') else None


    class DirEntry(KaitaiStruct):

        class FileTypeEnum(Enum):
            unknown = 0
            reg_file = 1
            dir = 2
            chrdev = 3
            blkdev = 4
            fifo = 5
            sock = 6
            symlink = 7
        SEQ_FIELDS = ["inode_ptr", "rec_len", "name_len", "file_type", "name", "padding"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['inode_ptr']['start'] = self._io.pos()
            self.inode_ptr = self._io.read_u4le()
            self._debug['inode_ptr']['end'] = self._io.pos()
            self._debug['rec_len']['start'] = self._io.pos()
            self.rec_len = self._io.read_u2le()
            self._debug['rec_len']['end'] = self._io.pos()
            self._debug['name_len']['start'] = self._io.pos()
            self.name_len = self._io.read_u1()
            self._debug['name_len']['end'] = self._io.pos()
            self._debug['file_type']['start'] = self._io.pos()
            self.file_type = KaitaiStream.resolve_enum(Ext2.DirEntry.FileTypeEnum, self._io.read_u1())
            self._debug['file_type']['end'] = self._io.pos()
            self._debug['name']['start'] = self._io.pos()
            self.name = (self._io.read_bytes(self.name_len)).decode(u"UTF-8")
            self._debug['name']['end'] = self._io.pos()
            self._debug['padding']['start'] = self._io.pos()
            self.padding = self._io.read_bytes(((self.rec_len - self.name_len) - 8))
            self._debug['padding']['end'] = self._io.pos()

        @property
        def inode(self):
            if hasattr(self, '_m_inode'):
                return self._m_inode if hasattr(self, '_m_inode') else None

            self._m_inode = self._root.bg1.block_groups[(self.inode_ptr - 1) // self._root.bg1.super_block.inodes_per_group].inodes[((self.inode_ptr - 1) % self._root.bg1.super_block.inodes_per_group)]
            return self._m_inode if hasattr(self, '_m_inode') else None


    class Inode(KaitaiStruct):
        SEQ_FIELDS = ["mode", "uid", "size", "atime", "ctime", "mtime", "dtime", "gid", "links_count", "blocks", "flags", "osd1", "block", "generation", "file_acl", "dir_acl", "faddr", "osd2"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['mode']['start'] = self._io.pos()
            self.mode = self._io.read_u2le()
            self._debug['mode']['end'] = self._io.pos()
            self._debug['uid']['start'] = self._io.pos()
            self.uid = self._io.read_u2le()
            self._debug['uid']['end'] = self._io.pos()
            self._debug['size']['start'] = self._io.pos()
            self.size = self._io.read_u4le()
            self._debug['size']['end'] = self._io.pos()
            self._debug['atime']['start'] = self._io.pos()
            self.atime = self._io.read_u4le()
            self._debug['atime']['end'] = self._io.pos()
            self._debug['ctime']['start'] = self._io.pos()
            self.ctime = self._io.read_u4le()
            self._debug['ctime']['end'] = self._io.pos()
            self._debug['mtime']['start'] = self._io.pos()
            self.mtime = self._io.read_u4le()
            self._debug['mtime']['end'] = self._io.pos()
            self._debug['dtime']['start'] = self._io.pos()
            self.dtime = self._io.read_u4le()
            self._debug['dtime']['end'] = self._io.pos()
            self._debug['gid']['start'] = self._io.pos()
            self.gid = self._io.read_u2le()
            self._debug['gid']['end'] = self._io.pos()
            self._debug['links_count']['start'] = self._io.pos()
            self.links_count = self._io.read_u2le()
            self._debug['links_count']['end'] = self._io.pos()
            self._debug['blocks']['start'] = self._io.pos()
            self.blocks = self._io.read_u4le()
            self._debug['blocks']['end'] = self._io.pos()
            self._debug['flags']['start'] = self._io.pos()
            self.flags = self._io.read_u4le()
            self._debug['flags']['end'] = self._io.pos()
            self._debug['osd1']['start'] = self._io.pos()
            self.osd1 = self._io.read_u4le()
            self._debug['osd1']['end'] = self._io.pos()
            self._debug['block']['start'] = self._io.pos()
            self.block = [None] * (15)
            for i in range(15):
                if not 'arr' in self._debug['block']:
                    self._debug['block']['arr'] = []
                self._debug['block']['arr'].append({'start': self._io.pos()})
                _t_block = Ext2.BlockPtr(self._io, self, self._root)
                _t_block._read()
                self.block[i] = _t_block
                self._debug['block']['arr'][i]['end'] = self._io.pos()

            self._debug['block']['end'] = self._io.pos()
            self._debug['generation']['start'] = self._io.pos()
            self.generation = self._io.read_u4le()
            self._debug['generation']['end'] = self._io.pos()
            self._debug['file_acl']['start'] = self._io.pos()
            self.file_acl = self._io.read_u4le()
            self._debug['file_acl']['end'] = self._io.pos()
            self._debug['dir_acl']['start'] = self._io.pos()
            self.dir_acl = self._io.read_u4le()
            self._debug['dir_acl']['end'] = self._io.pos()
            self._debug['faddr']['start'] = self._io.pos()
            self.faddr = self._io.read_u4le()
            self._debug['faddr']['end'] = self._io.pos()
            self._debug['osd2']['start'] = self._io.pos()
            self.osd2 = self._io.read_bytes(12)
            self._debug['osd2']['end'] = self._io.pos()

        @property
        def as_dir(self):
            if hasattr(self, '_m_as_dir'):
                return self._m_as_dir if hasattr(self, '_m_as_dir') else None

            io = self.block[0].body._io
            _pos = io.pos()
            io.seek(0)
            self._debug['_m_as_dir']['start'] = io.pos()
            self._m_as_dir = Ext2.Dir(io, self, self._root)
            self._m_as_dir._read()
            self._debug['_m_as_dir']['end'] = io.pos()
            io.seek(_pos)
            return self._m_as_dir if hasattr(self, '_m_as_dir') else None


    class BlockPtr(KaitaiStruct):
        SEQ_FIELDS = ["ptr"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['ptr']['start'] = self._io.pos()
            self.ptr = self._io.read_u4le()
            self._debug['ptr']['end'] = self._io.pos()

        @property
        def body(self):
            if hasattr(self, '_m_body'):
                return self._m_body if hasattr(self, '_m_body') else None

            _pos = self._io.pos()
            self._io.seek((self.ptr * self._root.bg1.super_block.block_size))
            self._debug['_m_body']['start'] = self._io.pos()
            self._raw__m_body = self._io.read_bytes(self._root.bg1.super_block.block_size)
            _io__raw__m_body = KaitaiStream(BytesIO(self._raw__m_body))
            self._m_body = Ext2.RawBlock(_io__raw__m_body, self, self._root)
            self._m_body._read()
            self._debug['_m_body']['end'] = self._io.pos()
            self._io.seek(_pos)
            return self._m_body if hasattr(self, '_m_body') else None


    class Dir(KaitaiStruct):
        SEQ_FIELDS = ["entries"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['entries']['start'] = self._io.pos()
            self.entries = []
            i = 0
            while not self._io.is_eof():
                if not 'arr' in self._debug['entries']:
                    self._debug['entries']['arr'] = []
                self._debug['entries']['arr'].append({'start': self._io.pos()})
                _t_entries = Ext2.DirEntry(self._io, self, self._root)
                _t_entries._read()
                self.entries.append(_t_entries)
                self._debug['entries']['arr'][len(self.entries) - 1]['end'] = self._io.pos()
                i += 1

            self._debug['entries']['end'] = self._io.pos()


    class BlockGroup(KaitaiStruct):
        SEQ_FIELDS = ["super_block", "block_groups"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['super_block']['start'] = self._io.pos()
            self._raw_super_block = self._io.read_bytes(1024)
            _io__raw_super_block = KaitaiStream(BytesIO(self._raw_super_block))
            self.super_block = Ext2.SuperBlockStruct(_io__raw_super_block, self, self._root)
            self.super_block._read()
            self._debug['super_block']['end'] = self._io.pos()
            self._debug['block_groups']['start'] = self._io.pos()
            self.block_groups = [None] * (self.super_block.block_group_count)
            for i in range(self.super_block.block_group_count):
                if not 'arr' in self._debug['block_groups']:
                    self._debug['block_groups']['arr'] = []
                self._debug['block_groups']['arr'].append({'start': self._io.pos()})
                _t_block_groups = Ext2.Bgd(self._io, self, self._root)
                _t_block_groups._read()
                self.block_groups[i] = _t_block_groups
                self._debug['block_groups']['arr'][i]['end'] = self._io.pos()

            self._debug['block_groups']['end'] = self._io.pos()


    class Bgd(KaitaiStruct):
        SEQ_FIELDS = ["block_bitmap_block", "inode_bitmap_block", "inode_table_block", "free_blocks_count", "free_inodes_count", "used_dirs_count", "pad_reserved"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['block_bitmap_block']['start'] = self._io.pos()
            self.block_bitmap_block = self._io.read_u4le()
            self._debug['block_bitmap_block']['end'] = self._io.pos()
            self._debug['inode_bitmap_block']['start'] = self._io.pos()
            self.inode_bitmap_block = self._io.read_u4le()
            self._debug['inode_bitmap_block']['end'] = self._io.pos()
            self._debug['inode_table_block']['start'] = self._io.pos()
            self.inode_table_block = self._io.read_u4le()
            self._debug['inode_table_block']['end'] = self._io.pos()
            self._debug['free_blocks_count']['start'] = self._io.pos()
            self.free_blocks_count = self._io.read_u2le()
            self._debug['free_blocks_count']['end'] = self._io.pos()
            self._debug['free_inodes_count']['start'] = self._io.pos()
            self.free_inodes_count = self._io.read_u2le()
            self._debug['free_inodes_count']['end'] = self._io.pos()
            self._debug['used_dirs_count']['start'] = self._io.pos()
            self.used_dirs_count = self._io.read_u2le()
            self._debug['used_dirs_count']['end'] = self._io.pos()
            self._debug['pad_reserved']['start'] = self._io.pos()
            self.pad_reserved = self._io.read_bytes((2 + 12))
            self._debug['pad_reserved']['end'] = self._io.pos()

        @property
        def block_bitmap(self):
            if hasattr(self, '_m_block_bitmap'):
                return self._m_block_bitmap if hasattr(self, '_m_block_bitmap') else None

            _pos = self._io.pos()
            self._io.seek((self.block_bitmap_block * self._root.bg1.super_block.block_size))
            self._debug['_m_block_bitmap']['start'] = self._io.pos()
            self._m_block_bitmap = self._io.read_bytes(1024)
            self._debug['_m_block_bitmap']['end'] = self._io.pos()
            self._io.seek(_pos)
            return self._m_block_bitmap if hasattr(self, '_m_block_bitmap') else None

        @property
        def inode_bitmap(self):
            if hasattr(self, '_m_inode_bitmap'):
                return self._m_inode_bitmap if hasattr(self, '_m_inode_bitmap') else None

            _pos = self._io.pos()
            self._io.seek((self.inode_bitmap_block * self._root.bg1.super_block.block_size))
            self._debug['_m_inode_bitmap']['start'] = self._io.pos()
            self._m_inode_bitmap = self._io.read_bytes(1024)
            self._debug['_m_inode_bitmap']['end'] = self._io.pos()
            self._io.seek(_pos)
            return self._m_inode_bitmap if hasattr(self, '_m_inode_bitmap') else None

        @property
        def inodes(self):
            if hasattr(self, '_m_inodes'):
                return self._m_inodes if hasattr(self, '_m_inodes') else None

            _pos = self._io.pos()
            self._io.seek((self.inode_table_block * self._root.bg1.super_block.block_size))
            self._debug['_m_inodes']['start'] = self._io.pos()
            self._m_inodes = [None] * (self._root.bg1.super_block.inodes_per_group)
            for i in range(self._root.bg1.super_block.inodes_per_group):
                if not 'arr' in self._debug['_m_inodes']:
                    self._debug['_m_inodes']['arr'] = []
                self._debug['_m_inodes']['arr'].append({'start': self._io.pos()})
                _t__m_inodes = Ext2.Inode(self._io, self, self._root)
                _t__m_inodes._read()
                self._m_inodes[i] = _t__m_inodes
                self._debug['_m_inodes']['arr'][i]['end'] = self._io.pos()

            self._debug['_m_inodes']['end'] = self._io.pos()
            self._io.seek(_pos)
            return self._m_inodes if hasattr(self, '_m_inodes') else None


    class RawBlock(KaitaiStruct):
        SEQ_FIELDS = ["body"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['body']['start'] = self._io.pos()
            self.body = self._io.read_bytes(self._root.bg1.super_block.block_size)
            self._debug['body']['end'] = self._io.pos()


    @property
    def bg1(self):
        if hasattr(self, '_m_bg1'):
            return self._m_bg1 if hasattr(self, '_m_bg1') else None

        _pos = self._io.pos()
        self._io.seek(1024)
        self._debug['_m_bg1']['start'] = self._io.pos()
        self._m_bg1 = Ext2.BlockGroup(self._io, self, self._root)
        self._m_bg1._read()
        self._debug['_m_bg1']['end'] = self._io.pos()
        self._io.seek(_pos)
        return self._m_bg1 if hasattr(self, '_m_bg1') else None

    @property
    def root_dir(self):
        if hasattr(self, '_m_root_dir'):
            return self._m_root_dir if hasattr(self, '_m_root_dir') else None

        self._m_root_dir = self.bg1.block_groups[0].inodes[1].as_dir
        return self._m_root_dir if hasattr(self, '_m_root_dir') else None





# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

from pkg_resources import parse_version
import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO
from enum import Enum
import collections


if parse_version(kaitaistruct.__version__) < parse_version('0.9'):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

from polyfile.kaitai.parsers import icmp_packet
from polyfile.kaitai.parsers import udp_datagram
from polyfile.kaitai.parsers import ipv4_packet
from polyfile.kaitai.parsers import ipv6_packet
from polyfile.kaitai.parsers import tcp_segment
class ProtocolBody(KaitaiStruct):
    """Protocol body represents particular payload on transport level (OSI
    layer 4).
    
    Typically this payload in encapsulated into network level (OSI layer
    3) packet, which includes "protocol number" field that would be used
    to decide what's inside the payload and how to parse it. Thanks to
    IANA's standardization effort, multiple network level use the same
    IDs for these payloads named "protocol numbers".
    
    This is effectively a "router" type: it expects to get protocol
    number as a parameter, and then invokes relevant type parser based
    on that parameter.
    
    .. seealso::
       Source - https://www.iana.org/assignments/protocol-numbers/protocol-numbers.xhtml
    """

    class ProtocolEnum(Enum):
        hopopt = 0
        icmp = 1
        igmp = 2
        ggp = 3
        ipv4 = 4
        st = 5
        tcp = 6
        cbt = 7
        egp = 8
        igp = 9
        bbn_rcc_mon = 10
        nvp_ii = 11
        pup = 12
        argus = 13
        emcon = 14
        xnet = 15
        chaos = 16
        udp = 17
        mux = 18
        dcn_meas = 19
        hmp = 20
        prm = 21
        xns_idp = 22
        trunk_1 = 23
        trunk_2 = 24
        leaf_1 = 25
        leaf_2 = 26
        rdp = 27
        irtp = 28
        iso_tp4 = 29
        netblt = 30
        mfe_nsp = 31
        merit_inp = 32
        dccp = 33
        x_3pc = 34
        idpr = 35
        xtp = 36
        ddp = 37
        idpr_cmtp = 38
        tp_plus_plus = 39
        il = 40
        ipv6 = 41
        sdrp = 42
        ipv6_route = 43
        ipv6_frag = 44
        idrp = 45
        rsvp = 46
        gre = 47
        dsr = 48
        bna = 49
        esp = 50
        ah = 51
        i_nlsp = 52
        swipe = 53
        narp = 54
        mobile = 55
        tlsp = 56
        skip = 57
        ipv6_icmp = 58
        ipv6_nonxt = 59
        ipv6_opts = 60
        any_host_internal_protocol = 61
        cftp = 62
        any_local_network = 63
        sat_expak = 64
        kryptolan = 65
        rvd = 66
        ippc = 67
        any_distributed_file_system = 68
        sat_mon = 69
        visa = 70
        ipcv = 71
        cpnx = 72
        cphb = 73
        wsn = 74
        pvp = 75
        br_sat_mon = 76
        sun_nd = 77
        wb_mon = 78
        wb_expak = 79
        iso_ip = 80
        vmtp = 81
        secure_vmtp = 82
        vines = 83
        ttp_or_iptm = 84
        nsfnet_igp = 85
        dgp = 86
        tcf = 87
        eigrp = 88
        ospfigp = 89
        sprite_rpc = 90
        larp = 91
        mtp = 92
        ax_25 = 93
        ipip = 94
        micp = 95
        scc_sp = 96
        etherip = 97
        encap = 98
        any_private_encryption_scheme = 99
        gmtp = 100
        ifmp = 101
        pnni = 102
        pim = 103
        aris = 104
        scps = 105
        qnx = 106
        a_n = 107
        ipcomp = 108
        snp = 109
        compaq_peer = 110
        ipx_in_ip = 111
        vrrp = 112
        pgm = 113
        any_0_hop = 114
        l2tp = 115
        ddx = 116
        iatp = 117
        stp = 118
        srp = 119
        uti = 120
        smp = 121
        sm = 122
        ptp = 123
        isis_over_ipv4 = 124
        fire = 125
        crtp = 126
        crudp = 127
        sscopmce = 128
        iplt = 129
        sps = 130
        pipe = 131
        sctp = 132
        fc = 133
        rsvp_e2e_ignore = 134
        mobility_header = 135
        udplite = 136
        mpls_in_ip = 137
        manet = 138
        hip = 139
        shim6 = 140
        wesp = 141
        rohc = 142
        reserved_255 = 255
    SEQ_FIELDS = ["body"]
    def __init__(self, protocol_num, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self.protocol_num = protocol_num
        self._debug = collections.defaultdict(dict)

    def _read(self):
        self._debug['body']['start'] = self._io.pos()
        _on = self.protocol
        if _on == ProtocolBody.ProtocolEnum.ipv6_nonxt:
            self.body = ProtocolBody.NoNextHeader(self._io, self, self._root)
            self.body._read()
        elif _on == ProtocolBody.ProtocolEnum.ipv4:
            self.body = ipv4_packet.Ipv4Packet(self._io)
            self.body._read()
        elif _on == ProtocolBody.ProtocolEnum.udp:
            self.body = udp_datagram.UdpDatagram(self._io)
            self.body._read()
        elif _on == ProtocolBody.ProtocolEnum.icmp:
            self.body = icmp_packet.IcmpPacket(self._io)
            self.body._read()
        elif _on == ProtocolBody.ProtocolEnum.hopopt:
            self.body = ProtocolBody.OptionHopByHop(self._io, self, self._root)
            self.body._read()
        elif _on == ProtocolBody.ProtocolEnum.ipv6:
            self.body = ipv6_packet.Ipv6Packet(self._io)
            self.body._read()
        elif _on == ProtocolBody.ProtocolEnum.tcp:
            self.body = tcp_segment.TcpSegment(self._io)
            self.body._read()
        self._debug['body']['end'] = self._io.pos()

    class NoNextHeader(KaitaiStruct):
        """Dummy type for IPv6 "no next header" type, which signifies end of headers chain."""
        SEQ_FIELDS = []
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            pass


    class OptionHopByHop(KaitaiStruct):
        SEQ_FIELDS = ["next_header_type", "hdr_ext_len", "body", "next_header"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['next_header_type']['start'] = self._io.pos()
            self.next_header_type = self._io.read_u1()
            self._debug['next_header_type']['end'] = self._io.pos()
            self._debug['hdr_ext_len']['start'] = self._io.pos()
            self.hdr_ext_len = self._io.read_u1()
            self._debug['hdr_ext_len']['end'] = self._io.pos()
            self._debug['body']['start'] = self._io.pos()
            self.body = self._io.read_bytes((self.hdr_ext_len - 1))
            self._debug['body']['end'] = self._io.pos()
            self._debug['next_header']['start'] = self._io.pos()
            self.next_header = ProtocolBody(self.next_header_type, self._io)
            self.next_header._read()
            self._debug['next_header']['end'] = self._io.pos()


    @property
    def protocol(self):
        if hasattr(self, '_m_protocol'):
            return self._m_protocol if hasattr(self, '_m_protocol') else None

        self._m_protocol = KaitaiStream.resolve_enum(ProtocolBody.ProtocolEnum, self.protocol_num)
        return self._m_protocol if hasattr(self, '_m_protocol') else None


